using System;
using System.IO;
using System.Threading;
using System.Collections;
using System.Diagnostics;
using Microsoft.Build.Framework;
using Microsoft.Build.Utilities;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.Build.Proxy;

namespace Callista.BuildLab.BuildTasks
{
    public class VB6Build : Task
    {
        private string _teamFoundationUrl;
        private string _buildUri;
        private string _compilerPath;
        private string _projectFile;
        private int _timeout;
        private string _output;
        private int _exitCode = 0;

        private Queue standardErrorData = new Queue();
        private Queue standardOutputData = new Queue();
        private ManualResetEvent toolExited = new ManualResetEvent(false);
        private ManualResetEvent toolTimeout = new ManualResetEvent(false);

        [Required]
        public string TeamFoundationUrl
        {
            get { return _teamFoundationUrl; }
            set { _teamFoundationUrl = value; }
        }

        [Required]
        public string BuildUri
        {
            get { return _buildUri; }
            set { _buildUri = value; }
        }
        
        [Required]
        public string CompilerPath
        {
            get { return _compilerPath; }
            set { _compilerPath = value; }
        }

        [Required]
        public string ProjectFile
        {
            get { return _projectFile; }
            set { _projectFile = value; }
        }

        [Required]
        public int Timeout
        {
            get { return _timeout; }
            set { _timeout = value; }
        }

        [Required]
        public string Output
        {
            get { return _output; }
            set { _output = value; }
        }

        [Output]
        public int ExitCode
        {
            get { return _exitCode; }
        }

        private string AddTrailingSlash(string source)
        {
            string retval = source;

            if (!retval.EndsWith("\\"))
                retval = retval + "\\";

            return retval;
        }

        private string PadWithQuotes(string source)
        {
            string retval = source;

            if (!retval.StartsWith("\""))
                retval = "\"" + retval;

            if (!retval.EndsWith("\""))
                retval = retval + "\"";

            return retval;
        }

        private string ExtractProjectFileName(string projectFilePath)
        {
            int extensionIndex = projectFilePath.IndexOf(".vbp");
            int prefixingSlashIndex = projectFilePath.LastIndexOf("\\", extensionIndex);

            return projectFilePath.Substring(prefixingSlashIndex + 1, (extensionIndex + 4) - (prefixingSlashIndex + 1));
        }
        
        private void ReceiveExitNotification(object sender, EventArgs e)
        {
            toolExited.Set();
        }

        private void ReceiveTimeoutNotification(object unused)
        {
            toolTimeout.Set();
        }

        private void ReceiveStandardErrorData(object sender, DataReceivedEventArgs e)
        {
            if (e.Data != null)
            {
                lock (standardErrorData.SyncRoot)
                {
                    standardErrorData.Enqueue(e.Data);
                }
            }
        }

        private void ReceiveStandardOutputData(object sender, DataReceivedEventArgs e)
        {
            if (e.Data != null)
            {
                lock (standardOutputData.SyncRoot)
                {
                    standardOutputData.Enqueue(e.Data);
                }
            }
        }

        private void HandleNotifications(Process p)
        {
            WaitHandle[] handles = new WaitHandle[] { toolTimeout, toolExited };
            switch (WaitHandle.WaitAny(handles))
            {
                case 0:
                    {
                        p.Kill();
                        break;
                    }
                case 1:
                    {
                        p.WaitForExit();
                        break;
                    }
            }

            try
            {
                _exitCode = p.ExitCode;
            }
            catch (InvalidOperationException)
            {
                _exitCode = -1;
            }
        }

        public override bool Execute()
        {
            Process p = null;
            Timer t = null;

            string projectFile = "";
            string projectFilePath = "";
            string outputFolder = "";
            string logFile = "";
            string logFilePath = "";

            try
            {
                standardErrorData = new Queue();
                standardOutputData = new Queue();
                toolExited = new ManualResetEvent(false);
                toolTimeout = new ManualResetEvent(false);

                TeamFoundationServer tfs = new TeamFoundationServer(TeamFoundationUrl);
                BuildStore bs = (BuildStore)(tfs.GetService(typeof(BuildStore)));

                p = new Process();

                projectFile = ExtractProjectFileName(ProjectFile);
                projectFilePath = PadWithQuotes(ProjectFile);
                
                outputFolder = PadWithQuotes(AddTrailingSlash(Output));
                
                logFile = projectFile.Replace(".vbp", ".log");
                logFilePath = AddTrailingSlash(Output) + logFile;

                string buildStepName = "CompilingVB6Project";
                string buildStepMsg = "Compiling " + projectFile;
                string buildResults = "";

                bs.AddBuildStep(BuildUri, buildStepName, buildStepMsg);

                Log.LogMessage("Compilation started: " + projectFile);

                p.StartInfo.CreateNoWindow = true;
                p.StartInfo.UseShellExecute = false;
                p.StartInfo.RedirectStandardError = true;
                p.StartInfo.RedirectStandardOutput = true;
                p.StartInfo.RedirectStandardInput = true;
                p.StartInfo.FileName = CompilerPath;
                p.StartInfo.Arguments = "/m " + projectFilePath + " /outdir " + outputFolder + " /out " + PadWithQuotes(logFilePath);
                p.StartInfo.WorkingDirectory = Environment.GetFolderPath(Environment.SpecialFolder.System);

                p.EnableRaisingEvents = true;
                p.Exited += new EventHandler(this.ReceiveExitNotification);
                p.ErrorDataReceived += new DataReceivedEventHandler(this.ReceiveStandardErrorData);
                p.OutputDataReceived += new DataReceivedEventHandler(this.ReceiveStandardOutputData);

                Log.LogMessage("\tCompiler Path: " + p.StartInfo.FileName);
                Log.LogMessage("\tCompiler Arguments: " + p.StartInfo.Arguments);

                p.Start();

                p.StandardInput.Close();
                p.BeginErrorReadLine();
                p.BeginOutputReadLine();

                t = new Timer(new TimerCallback(this.ReceiveTimeoutNotification), null, Timeout, -1);

                HandleNotifications(p);

                buildResults = File.ReadAllText(logFilePath);

                if (!buildResults.EndsWith("succeeded.\r\n"))
                    _exitCode = -1;

                if (_exitCode == 0)
                    bs.UpdateBuildStep(BuildUri, buildStepName, DateTime.Now, BuildStepStatus.Succeeded);
                else
                    bs.UpdateBuildStep(BuildUri, buildStepName, DateTime.Now, BuildStepStatus.Failed);

                Log.LogMessage("Standard Output:\n");
                foreach (string s in standardOutputData)
                    Log.LogMessage("\n" + s);

                Log.LogMessage("Standard Error:\n");
                foreach (string s in standardErrorData)
                    Log.LogMessage("\n" + s);

                Log.LogMessage("Compiler Log:" + buildResults);
                Log.LogMessage("Compilation completed: " + ProjectFile);
            }
            catch (Exception ex)
            {
                EventLog.WriteEntry("TfsBuildLab VB6Build", ex.ToString(), EventLogEntryType.Error);
                _exitCode = -1;
            }
            finally
            {
                File.Delete(logFilePath);

                if(toolExited != null)
                    toolExited.Close();
                if(toolTimeout != null)
                    toolTimeout.Close();

                if (t != null)
                    t.Dispose();
                if (p != null)
                    p.Close();
            }

            if (_exitCode == 0 && standardErrorData.Count == 0)
                return true;
            else
                return false;
        }
    }
}
