using System;
using System.Text;
using System.Data;
using System.Data.SqlClient;
using System.Collections.Generic;
using Callista.BuildLab.Core.Model;

namespace Callista.BuildLab.Core.Data
{
    public interface IBuildQueueRepository
    {
        void PushBuildToQueue(QueueEntry entry);
        QueueEntry PeekNextBuildOnQueue(long id);
        QueueEntry PopBuildFromQueue(long id);
        void IncrementBuildDeliveryAttempts(long id);
        long QueueEntryCount();
    }
    
    public interface IBuildQueueRepositoryAdmin
    {
        List<QueueEntry> ListQueuedBuilds(string teamProject, string buildType);
        void MoveBuildToBeginingOfQueue(int id);
        void MoveBuildToEndOfQueue(int id);
        void RemoveBuildFromQueue(int id);
    }

    public class BuildQueueRepository : IBuildQueueRepository, IBuildQueueRepositoryAdmin
    {
        #region IBuildQueueRepositoryAdmin Members

        List<QueueEntry> IBuildQueueRepositoryAdmin.ListQueuedBuilds(string teamProject, string buildType)
        {
            List<QueueEntry> qeList = new List<QueueEntry>();

            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                using (SqlCommand cmd = new SqlCommand("dbo.Queue_List", con))
                {
                    cmd.Parameters.Add("@TeamProject", SqlDbType.NVarChar, 128);
                    cmd.Parameters.Add("@BuildType", SqlDbType.NVarChar, 128);

                    if (teamProject != string.Empty)
                    {
                        cmd.Parameters["@TeamProject"].Value = teamProject;
                    }
                    else
                    {
                        cmd.Parameters["@TeamProject"].Value = DBNull.Value;
                    }

                    if (buildType != string.Empty)
                    {
                        cmd.Parameters["@BuildType"].Value = buildType;
                    }
                    else
                    {
                        cmd.Parameters["@BuildType"].Value = DBNull.Value;
                    }
                    
                    cmd.CommandType = CommandType.StoredProcedure;

                    using (SqlDataReader rdr = cmd.ExecuteReader())
                    {
                        while (rdr.Read())
                        {
                            int queueID = rdr.GetInt32(0);
                            string projectName = rdr.GetString(1);
                            string buildTypeName = rdr.GetString(2);
                            string buildServer = rdr.GetString(3);
                            Nullable<int> triggeringChangeset = rdr.IsDBNull(4) ? new Nullable<int>() : new Nullable<int>(rdr.GetInt32(4));
                            string triggeringUser = rdr.IsDBNull(5) ? "" : rdr.GetString(5);
                            string overriddenParameters = rdr.IsDBNull(6) ? "" : rdr.GetString(6);

                            qeList.Add(new QueueEntry(queueID, projectName, buildTypeName, buildServer, triggeringUser, triggeringChangeset, Common.DeserializeBuildParams(overriddenParameters)));
                        }
                    }
                }
            }

            return qeList;
        }

        void IBuildQueueRepositoryAdmin.MoveBuildToBeginingOfQueue(int id)
        {
            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                using (SqlCommand cmd = new SqlCommand("dbo.Queue_MoveFirst", con))
                {
                    cmd.Parameters.Add("@ID", SqlDbType.Int);
                    cmd.Parameters["@ID"].Value = id;

                    cmd.CommandType = CommandType.StoredProcedure;
                    cmd.ExecuteNonQuery();
                }
            }
        }

        void IBuildQueueRepositoryAdmin.MoveBuildToEndOfQueue(int id)
        {
            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                using (SqlCommand cmd = new SqlCommand("dbo.Queue_MoveLast", con))
                {
                    cmd.Parameters.Add("@ID", SqlDbType.Int);
                    cmd.Parameters["@ID"].Value = id;

                    cmd.CommandType = CommandType.StoredProcedure;
                    cmd.ExecuteNonQuery();
                }
            }
        }

        void IBuildQueueRepositoryAdmin.RemoveBuildFromQueue(int id)
        {
            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                using (SqlCommand cmd = new SqlCommand("dbo.Queue_Remove", con))
                {
                    cmd.Parameters.Add("@ID", SqlDbType.Int);
                    cmd.Parameters["@ID"].Value = id;

                    cmd.CommandType = CommandType.StoredProcedure;
                    cmd.ExecuteNonQuery();
                }
            }
        }

        #endregion

        #region IBuildQueueRepository Members

        void IBuildQueueRepository.PushBuildToQueue(QueueEntry entry)
        {
            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                using (SqlCommand cmd = new SqlCommand("dbo.Queue_Push", con))
                {
                    cmd.Parameters.Add("@TeamProject", SqlDbType.NVarChar, 128);
                    cmd.Parameters.Add("@BuildType", SqlDbType.NVarChar, 128);
                    cmd.Parameters.Add("@BuildServer", SqlDbType.NVarChar, 128);
                    cmd.Parameters.Add("@Changeset", SqlDbType.Int);
                    cmd.Parameters.Add("@Owner", SqlDbType.NVarChar, 128);
                    cmd.Parameters.Add("@BuildParams", SqlDbType.NText);

                    cmd.Parameters["@TeamProject"].Value = entry.TeamProjectName;
                    cmd.Parameters["@BuildType"].Value = entry.BuildTypeName;
                    cmd.Parameters["@BuildServer"].Value = entry.BuildServer;

                    if (entry.TriggeringChangesetID.HasValue)
                    {
                        cmd.Parameters["@Changeset"].Value = entry.TriggeringChangesetID.Value;
                    }
                    else
                    {
                        cmd.Parameters["@Changeset"].Value = DBNull.Value;
                    }

                    if (entry.TriggeringUser != string.Empty)
                    {
                        cmd.Parameters["@Owner"].Value = entry.TriggeringUser;
                    }
                    else
                    {
                        cmd.Parameters["@Owner"].Value = DBNull.Value;
                    }

                    if (entry.OverriddenBuildParameters.Count > 0)
                    {
                        cmd.Parameters["@BuildParams"].Value = Common.SerializeBuildParams(entry.OverriddenBuildParameters);
                    }
                    else
                    {
                        cmd.Parameters["@BuildParams"].Value = DBNull.Value;
                    }

                    cmd.CommandType = CommandType.StoredProcedure;
                    cmd.ExecuteNonQuery();
                }
            }
        }

        QueueEntry IBuildQueueRepository.PeekNextBuildOnQueue(long id)
        {
            QueueEntry qe = null;

            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                using (SqlCommand cmd = new SqlCommand("dbo.Queue_Peek", con))
                {
                    cmd.Parameters.Add("@ID", SqlDbType.Int);

                    if (id > -1)
                    {
                        cmd.Parameters["@ID"].Value = id;
                    }
                    else
                    {
                        cmd.Parameters["@ID"].Value = DBNull.Value;
                    }

                    cmd.CommandType = CommandType.StoredProcedure;

                    using (SqlDataReader rdr = cmd.ExecuteReader())
                    {
                        if (rdr.Read())
                        {
                            int queueID = rdr.GetInt32(0);
                            string projectName = rdr.GetString(1);
                            string buildTypeName = rdr.GetString(2);
                            string buildServer = rdr.GetString(3);
                            Nullable<int> triggeringChangeset = rdr.IsDBNull(4) ? new Nullable<int>() : new Nullable<int>(rdr.GetInt32(4));
                            string triggeringUser = rdr.IsDBNull(5) ? "" : rdr.GetString(5);
                            string overriddenParameters = rdr.IsDBNull(6) ? "" : rdr.GetString(6);

                            qe = new QueueEntry(queueID, projectName, buildTypeName, buildServer, triggeringUser, triggeringChangeset, Common.DeserializeBuildParams(overriddenParameters));
                        }
                    }
                }
            }

            return qe;
        }

        QueueEntry IBuildQueueRepository.PopBuildFromQueue(long id)
        {
            QueueEntry qe = null;

            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                using (SqlCommand cmd = new SqlCommand("dbo.Queue_Pop", con))
                {
                    cmd.Parameters.Add("@ID", SqlDbType.Int);

                    if (id > -1)
                    {
                        cmd.Parameters["@ID"].Value = id;
                    }
                    else
                    {
                        cmd.Parameters["@ID"].Value = DBNull.Value;
                    }

                    cmd.CommandType = CommandType.StoredProcedure;

                    using (SqlDataReader rdr = cmd.ExecuteReader())
                    {
                        if (rdr.Read())
                        {
                            int queueID = rdr.GetInt32(0);
                            string projectName = rdr.GetString(1);
                            string buildTypeName = rdr.GetString(2);
                            string buildServer = rdr.GetString(3);
                            Nullable<int> triggeringChangeset = rdr.IsDBNull(4) ? new Nullable<int>() : new Nullable<int>(rdr.GetInt32(4));
                            string triggeringUser = rdr.IsDBNull(5) ? "" : rdr.GetString(5);
                            string overriddenParameters = rdr.IsDBNull(6) ? "" : rdr.GetString(6);

                            qe = new QueueEntry(queueID, projectName, buildTypeName, buildServer, triggeringUser, triggeringChangeset, Common.DeserializeBuildParams(overriddenParameters));
                        }
                    }
                }
            }

            return qe;
        }

        void IBuildQueueRepository.IncrementBuildDeliveryAttempts(long id)
        {
            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                using (SqlCommand cmd = new SqlCommand("dbo.Queue_IncrementDeliveryAttempt", con))
                {
                    cmd.Parameters.Add("@ID", SqlDbType.Int);
                    cmd.Parameters["@ID"].Value = id;

                    cmd.CommandType = CommandType.StoredProcedure;
                    cmd.ExecuteNonQuery();
                }
            }
        }

        long IBuildQueueRepository.QueueEntryCount()
        {
            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                using (SqlCommand cmd = new SqlCommand("dbo.Queue_Count", con))
                {
                    cmd.CommandType = CommandType.StoredProcedure;
                    using (SqlDataReader rdr = cmd.ExecuteReader())
                    {
                        if (rdr.Read())
                        {
                            return rdr.GetInt32(0);
                        }
                    }
                }
            }

            return 0;
        }

        #endregion
    }
}
