using System;
using System.Text;
using System.Data;
using System.Data.SqlClient;
using System.Collections.Generic;
using Callista.BuildLab.Core.Model;

namespace Callista.BuildLab.Core.Data
{
    public interface IBuildTypeRepository
    {
        TeamProject GetTeamProject(string teamProjectName);
        void CopyBuildType(string sourceTeamProjectName, string sourceBuildTypeName, string targetTeamProjectName, string targetBuildTypeName);
        void AddRetentionPolicyToBuildType(string teamProject, string buildType, BuildStatus buildStatus, BuildQuality buildQuality, RetentionPolicyType type, int quantity);
        void UpdateRetentionPolicy(RetentionPolicy policy);
        void DeleteRetentionPolicy(int id);
        void AddTriggerToBuildType(string teamProject, string buildType, Trigger trigger);
        void UpdateTrigger(Trigger trigger);
        void DeleteTrigger(int id);
    }

    public class BuildTypeRepository : IBuildTypeRepository
    {
        private void PopulateBuildTypeInfoForProject(TeamProject project)
        {
            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                // Retreive retention information...
                using (SqlCommand cmd = new SqlCommand("dbo.BuildType_Get", con))
                {
                    cmd.Parameters.Add("@TeamProject", SqlDbType.NVarChar, 128);
                    cmd.Parameters["@TeamProject"].Value = project.Name;
                    cmd.CommandType = CommandType.StoredProcedure;

                    using (SqlDataReader rdr = cmd.ExecuteReader())
                    {
                        BuildType buildType = null;
                        RetentionPolicy policy = null;
                        Trigger trigger = null;

                        //Get Retention Policies
                        while(rdr.Read())
                        {
                            string buildTypeName = rdr.GetString(0);
                            int policyID = rdr.GetInt32(1);
                            BuildStatus buildStatus = (BuildStatus)Enum.Parse(typeof(BuildStatus), rdr.GetString(2), true);
                            BuildQuality buildQuality = (BuildQuality)Enum.Parse(typeof(BuildQuality), rdr.GetString(3), true);
                            RetentionPolicyType policyType = (RetentionPolicyType)Enum.Parse(typeof(RetentionPolicyType), rdr.GetString(4), true);
                            Nullable<int> policyQualtity = rdr.IsDBNull(4) ?  new Nullable<int>() : new Nullable<int>(rdr.GetInt32(5));

                            if (project.BuildTypes.ContainsKey(buildTypeName))
                            {
                                buildType = project.BuildTypes[buildTypeName];
                            }
                            else {
                                buildType = new BuildType(project.Name, buildTypeName);
                                project.BuildTypes.Add(buildTypeName, buildType);
                            }

                            if (buildType.RetentionPolicies.ContainsKey(policyID))
                            {
                                policy = buildType.RetentionPolicies[policyID];
                            }
                            else
                            {
                                policy = new RetentionPolicy(buildType, policyID, buildStatus, buildQuality, policyType, policyQualtity);
                                buildType.RetentionPolicies.Add(policyID, policy);
                            }
                        }

                        rdr.NextResult();

                        //Get Build Triggers
                        while (rdr.Read())
                        {
                            string buildTypeName = rdr.GetString(0);
                            int triggerID = rdr.GetInt32(1);
                            TriggerType triggerType = (TriggerType)Enum.Parse(typeof(TriggerType), rdr.GetString(2), true);
                            string triggerData = rdr.GetString(3);
                            string triggerParameters = rdr.IsDBNull(4) ?  "" : rdr.GetString(4);

                            if (project.BuildTypes.ContainsKey(buildTypeName))
                            {
                                buildType = project.BuildTypes[buildTypeName];
                            }
                            else
                            {
                                buildType = new BuildType(project.Name, buildTypeName);
                                project.BuildTypes.Add(buildTypeName, buildType);
                            }

                            if (buildType.Triggers.ContainsKey(triggerID))
                            {
                                trigger = buildType.Triggers[triggerID];
                            }
                            else
                            {
                                trigger = new Trigger(buildType, triggerID, triggerType, Common.ExtractTriggerPath(triggerData), Common.DeserializeBuildServers(triggerData), Common.DeserializeBuildParams(triggerParameters));
                                buildType.Triggers.Add(triggerID, trigger);
                            }
                        }
                    }
                }
            }
        }

        #region IBuildTypeRepository Members

        TeamProject IBuildTypeRepository.GetTeamProject(string teamProjectName)
        {
            TeamProject project = new TeamProject(teamProjectName);

            PopulateBuildTypeInfoForProject(project);

            return project;
        }

        void IBuildTypeRepository.CopyBuildType(string sourceTeamProjectName, string sourceBuildTypeName, string targetTeamProjectName, string targetBuildTypeName)
        {
            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                using (SqlCommand cmd = new SqlCommand("dbo.BuildType_CopyConfig", con))
                {
                    cmd.Parameters.Add("@SrcTeamProject", SqlDbType.NVarChar, 128);
                    cmd.Parameters.Add("@SrcBuildType", SqlDbType.NVarChar, 128);
                    cmd.Parameters.Add("@DstTeamProject", SqlDbType.NVarChar, 128);
                    cmd.Parameters.Add("@DstBuildType", SqlDbType.NVarChar, 128);

                    cmd.Parameters["@SrcTeamProject"].Value = sourceTeamProjectName;
                    cmd.Parameters["@SrcBuildType"].Value = sourceBuildTypeName;
                    cmd.Parameters["@DstTeamProject"].Value = targetTeamProjectName;
                    cmd.Parameters["@DstBuildType"].Value = targetBuildTypeName;

                    cmd.CommandType = CommandType.StoredProcedure;
                    cmd.ExecuteNonQuery();
                }
            }
        }

        void IBuildTypeRepository.AddRetentionPolicyToBuildType(string teamProject, string buildType, BuildStatus buildStatus, BuildQuality buildQuality, RetentionPolicyType type, int quantity)
        {
            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                using (SqlCommand cmd = new SqlCommand("dbo.BuildType_AddRetentionPolicy", con))
                {
                    cmd.Parameters.Add("@TeamProject", SqlDbType.NVarChar, 128);
                    cmd.Parameters.Add("@BuildType", SqlDbType.NVarChar, 128);
                    cmd.Parameters.Add("@BuildStatus", SqlDbType.NVarChar, 16);
                    cmd.Parameters.Add("@BuildQuality", SqlDbType.NVarChar, 32);
                    cmd.Parameters.Add("@PolicyType", SqlDbType.NVarChar, 16);
                    cmd.Parameters.Add("@Quantity", SqlDbType.Int);

                    cmd.Parameters["@TeamProject"].Value = teamProject;
                    cmd.Parameters["@BuildType"].Value = buildType;
                    cmd.Parameters["@BuildStatus"].Value = buildStatus;
                    cmd.Parameters["@BuildQuality"].Value = buildQuality;
                    cmd.Parameters["@PolicyType"].Value = type;
                    cmd.Parameters["@Quantity"].Value = quantity;

                    cmd.CommandType = CommandType.StoredProcedure;
                    cmd.ExecuteNonQuery();
                }
            }
        }

        void IBuildTypeRepository.AddTriggerToBuildType(string teamProject, string buildType, Trigger trigger)
        {
            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                using (SqlCommand cmd = new SqlCommand("dbo.BuildType_AddTrigger", con))
                {
                    cmd.Parameters.Add("@TeamProject", SqlDbType.NVarChar, 128);
                    cmd.Parameters.Add("@BuildType", SqlDbType.NVarChar, 128);
                    cmd.Parameters.Add("@TriggerType", SqlDbType.NVarChar, 16);
                    cmd.Parameters.Add("@Data", SqlDbType.NVarChar, 256);
                    cmd.Parameters.Add("@Parameters", SqlDbType.NText);

                    cmd.Parameters["@TeamProject"].Value = teamProject;
                    cmd.Parameters["@BuildType"].Value = buildType;
                    cmd.Parameters["@TriggerType"].Value = trigger.Type;
                    cmd.Parameters["@Data"].Value = trigger.Data + "|" + Common.SerializeBuildServers(trigger.Servers);
                    cmd.Parameters["@Parameters"].Value = Common.SerializeBuildParams(trigger.Parameters);

                    cmd.CommandType = CommandType.StoredProcedure;
                    cmd.ExecuteNonQuery();
                }
            }
        }

        void IBuildTypeRepository.UpdateRetentionPolicy(RetentionPolicy policy)
        {
            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                using (SqlCommand cmd = new SqlCommand("dbo.BuildType_UpdateRetentionPolicy", con))
                {
                    cmd.Parameters.Add("@ID", SqlDbType.Int);
                    cmd.Parameters.Add("@BuildStatus", SqlDbType.NVarChar, 16);
                    cmd.Parameters.Add("@BuildQuality", SqlDbType.NVarChar, 32);
                    cmd.Parameters.Add("@PolicyType", SqlDbType.NVarChar, 16);
                    cmd.Parameters.Add("@Quantity", SqlDbType.Int);

                    cmd.Parameters["@ID"].Value = policy.ID;
                    cmd.Parameters["@BuildStatus"].Value = policy.BuildStatus;
                    cmd.Parameters["@BuildQuality"].Value = policy.BuildQuality;
                    cmd.Parameters["@PolicyType"].Value = policy.Type;
                    cmd.Parameters["@Quantity"].Value = policy.RetentionQuantity;

                    cmd.CommandType = CommandType.StoredProcedure;
                    cmd.ExecuteNonQuery();
                }
            }
        }

        void IBuildTypeRepository.UpdateTrigger(Trigger trigger)
        {
            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                using (SqlCommand cmd = new SqlCommand("dbo.BuildType_UpdateTrigger", con))
                {
                    cmd.Parameters.Add("@ID", SqlDbType.Int);
                    cmd.Parameters.Add("@TriggerType", SqlDbType.NVarChar, 16);
                    cmd.Parameters.Add("@Data", SqlDbType.NVarChar, 256);
                    cmd.Parameters.Add("@Parameters", SqlDbType.NText);

                    cmd.Parameters["@ID"].Value = trigger.ID;
                    cmd.Parameters["@TriggerType"].Value = trigger.Type;
                    cmd.Parameters["@Data"].Value = trigger.Data + "|" + Common.SerializeBuildServers(trigger.Servers);
                    cmd.Parameters["@Parameters"].Value = Common.SerializeBuildParams(trigger.Parameters);

                    cmd.CommandType = CommandType.StoredProcedure;
                    cmd.ExecuteNonQuery();
                }
            }
        }

        void IBuildTypeRepository.DeleteRetentionPolicy(int id)
        {
            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                using (SqlCommand cmd = new SqlCommand("dbo.BuildType_DeleteRetentionPolicy", con))
                {
                    cmd.Parameters.Add("@ID", SqlDbType.Int);
                    cmd.Parameters["@ID"].Value = id;

                    cmd.CommandType = CommandType.StoredProcedure;
                    cmd.ExecuteNonQuery();
                }
            }
        }

        void IBuildTypeRepository.DeleteTrigger(int id)
        {
            using (SqlConnection con = new SqlConnection(Common.GetConnectionString()))
            {
                con.Open();

                using (SqlCommand cmd = new SqlCommand("dbo.BuildType_DeleteTrigger", con))
                {
                    cmd.Parameters.Add("@ID", SqlDbType.Int);
                    cmd.Parameters["@ID"].Value = id;

                    cmd.CommandType = CommandType.StoredProcedure;
                    cmd.ExecuteNonQuery();
                }
            }
        }

        #endregion
    }
}
