using System;
using System.Text;
using System.Text.RegularExpressions;
using System.Collections.Generic;
using System.Runtime.Serialization;
using System.ServiceModel;

namespace Callista.BuildLab.Core.Model
{
    [DataContract]
    public class BuildType
    {
        private string _buildTypeName;
        private string _teamProjectName;
        private string _defaultBuildDirectory;
        private string _defaultBuildServer;
        private List<BuildParameter> _parameters = new List<BuildParameter>();
        private Dictionary<int, Trigger> _triggers = new Dictionary<int, Trigger>();
        private Dictionary<int, RetentionPolicy> _retentionPolicies = new Dictionary<int, RetentionPolicy>();
        private WorkspaceMapping _workspaceMapping = new WorkspaceMapping();

        [DataMember]
        public string Name
        {
            get { return _buildTypeName; }
            set { _buildTypeName = value; }
        }

        [DataMember]
        public WorkspaceMapping WorkspaceMapping
        {
            get { return _workspaceMapping; }
            set { _workspaceMapping = value; }
        }

        public string DefaultBuildDirectory
        {
            get { return _defaultBuildDirectory; }
            set { _defaultBuildDirectory = value; }
        }

        public string DefaultBuildServer
        {
            get { return _defaultBuildServer; }
            set { _defaultBuildServer = value; }
        }

        [DataMember]
        public string TeamProjectName
        {
            get { return _teamProjectName; }
            set { _teamProjectName = value; }
        }

        [DataMember]
        public List<BuildParameter> Parameters
        {
            get { return _parameters; }
            set { _parameters = value; }
        }

        [DataMember]
        public Dictionary<int, Trigger> Triggers
        {
            get { return _triggers; }
            set { _triggers = value; }
        }

        [DataMember]
        public Dictionary<int, RetentionPolicy> RetentionPolicies
        {
            get { return _retentionPolicies; }
            set { _retentionPolicies = value; }
        }

        public BuildType(string teamProjectName, string buildTypeName)
        {
            this.TeamProjectName = teamProjectName;
            this.Name = buildTypeName;
        }

        public bool HasMatchingTrigger(Changeset cs)
        {
            int countIncludedItems = 0;
            int countExcludedItems = 0;

            foreach (VersionedItem vi in cs.VersionedItems)
            {
                foreach (Trigger t in Triggers.Values)
                {
                    if (t.Type != TriggerType.Schedule)
                    {
                        if (Regex.IsMatch(vi.FullPath, t.Data))
                        {
                            switch (t.Type)
                            {
                                case TriggerType.CheckinInclude:
                                    countIncludedItems++;
                                    break;
                                case TriggerType.CheckinExclude:
                                    countExcludedItems++;
                                    break;
                            }
                        }
                    }
                }
            }

            if (countIncludedItems > 0 && countExcludedItems < cs.VersionedItems.Count)
            {
                return true;
            }

            return false;
        }

        public List<string> GetAvailableBuildServers(List<Changeset> changesets)
        {
            List<string> result = new List<string>();

            Dictionary<int, Trigger> triggers = new Dictionary<int, Trigger>();
            Dictionary<string, string> servers = new Dictionary<string, string>();

            foreach (Changeset cs in changesets)
            {
                foreach (VersionedItem vi in cs.VersionedItems)
                {
                    foreach (Trigger t in Triggers.Values)
                    {
                        if (t.Type != TriggerType.Schedule)
                        {
                            if (Regex.IsMatch(vi.FullPath, t.Data))
                            {
                                if( t.Type == TriggerType.CheckinInclude)
                                {
                                    if (!triggers.ContainsKey(t.ID))
                                    {
                                        triggers.Add(t.ID, t);
                                    }
                                }
                            }
                        }
                    }
                }
            }

            //Always add the default build server first
            if (!servers.ContainsKey(this.DefaultBuildServer))
            {
                servers.Add(this.DefaultBuildServer, this.DefaultBuildServer);
            }

            foreach (Trigger t in triggers.Values)
            {
                foreach (string s in t.Servers)
                {
                    if (!servers.ContainsKey(s))
                    {
                        servers.Add(s,s);
                    }
                }
            }

            foreach (string server in servers.Values)
            {
                result.Add(server);
            }

            return result;
        }

        public Dictionary<string, string> GetBuildParametersForMatchingTriggers(Changeset cs)
        {
            Dictionary<string, string> parameters = new Dictionary<string, string>();

            //Add static parameters always available
            parameters.Add("ChangesetID", cs.Id.ToString());

            foreach (Trigger t in Triggers.Values)
            {
                if (t.Type == TriggerType.CheckinInclude)
                {
                    foreach (VersionedItem vi in cs.VersionedItems)
                    {
                        if (Regex.IsMatch(vi.FullPath, t.Data))
                        {
                            foreach (string key in t.Parameters.Keys)
                            {
                                if (parameters.ContainsKey(key))
                                {
                                    // Allways use the latest found value
                                    parameters[key] = t.Parameters[key];
                                }
                                else
                                {
                                    parameters.Add(key, t.Parameters[key]);
                                }
                            }
                            break;
                        }
                    }
                }
            }

            return parameters;
        }
    }
}
