using System;
using System.IO;
using System.Text;
using System.Diagnostics;
using System.Configuration;
using System.Collections.Generic;
using System.Runtime.Serialization;
using System.ServiceModel;
using Microsoft.TeamFoundation.Build;
using Microsoft.TeamFoundation.Build.Proxy;

namespace Callista.BuildLab.Core.Model
{
    [DataContract]
    public class RetentionPolicy
    {
        private BuildType _parent;
        private int _id;
        private BuildStatus _buildStatus = BuildStatus.Any;
        private BuildQuality _buildQuality = BuildQuality.Any;
        private RetentionPolicyType _retentionType;
        private Nullable<int> _retentionQuantity;

        public delegate bool DeleteBuildDelegate(string buildUri, out string failureMessage);
        public delegate void WriteLogEntryDelegate(EventType eventType, string teamProjectName, string buildTypeName, string triggeredBy, string buildUri, string message);

	    public bool OnlyCleanDropLocation
	    {
		    get 
            {
                string temp = ConfigurationManager.AppSettings["RetentionPoliciesOnlyCleanupDropLocation"];

                if (string.IsNullOrEmpty(temp))
                {
                    temp = "false";
                }

                return bool.Parse(temp);
            }
	    }
	
        public BuildType Parent
        {
            get { return _parent; }
        }

        [DataMember]
        public int ID
        {
            get { return _id; }
            set { _id = value; }
        }

        [DataMember]
        public BuildStatus BuildStatus
        {
            get { return _buildStatus; }
            set { _buildStatus = value; }
        }

        [DataMember]
        public BuildQuality BuildQuality
        {
            get { return _buildQuality; }
            set { _buildQuality = value; }
        }

        [DataMember]
        public RetentionPolicyType Type
        {
            get { return _retentionType; }
            set { _retentionType = value; }
        }

        [DataMember]
        public Nullable<int> RetentionQuantity
        {
            get { return _retentionQuantity; }
            set { _retentionQuantity = value; }
        }

        public RetentionPolicy(BuildType parent, int id, BuildStatus buildStatus, BuildQuality buildQuality, RetentionPolicyType retentionType, Nullable<int> retentionQuantity)
        {
            this._parent = parent;
            this._id = id;
            this._buildStatus = buildStatus;
            this._buildQuality = buildQuality;
            this._retentionType = retentionType;
            this._retentionQuantity = retentionQuantity;
        }

        public void EnforceRetentionPolicy(List<TeamBuild> builds, DeleteBuildDelegate buildDeleteFunction, WriteLogEntryDelegate writeLogFunction)
        {
            SortedList<DateTime, TeamBuild> result = new SortedList<DateTime, TeamBuild>();

            foreach (TeamBuild build in builds)
            {
                if (build.Status == BuildStatus || BuildStatus == BuildStatus.Any)
                {
                    if (build.Quality == BuildQuality || BuildQuality == BuildQuality.Any)
                    {
                        if (OnlyCleanDropLocation)
                        {
                            if (Directory.Exists(build.DropLocation))
                            {
                                result.Add(build.FinishTime, build);
                            }
                        }
                        else
                        {
                            result.Add(build.FinishTime, build);
                        }
                    }
                }
            }

            if (result.Count > 0)
            {
                switch (Type)
                {
                    case RetentionPolicyType.Builds:
                        EnforceNumberOfBuildsPolicy(result, buildDeleteFunction, writeLogFunction);
                        break;
                    case RetentionPolicyType.Days:
                        EnforceNumberOfDaysPolicy(result, buildDeleteFunction, writeLogFunction);
                        break;
                }
            }
        }

        private void EnforceNumberOfDaysPolicy(SortedList<DateTime, TeamBuild> result, DeleteBuildDelegate buildDeleteFunction, WriteLogEntryDelegate writeLogFunction)
        {
            DateTime start = DateTime.Now;
            string failure = "";

            while (result.Keys[0] < start.AddDays(-((double)RetentionQuantity)))
            {
                try 
	            {
                    if (OnlyCleanDropLocation)
                    {
                        if (Directory.Exists(result.Values[0].DropLocation))
                        {
                            Directory.Delete(result.Values[0].DropLocation, true);
                            writeLogFunction.Invoke(EventType.Delete, Parent.TeamProjectName, result.Values[0].BuildType, "", result.Values[0].BuildUri, "");
                        }
                    }
                    else
                    {
                        buildDeleteFunction.Invoke(result.Values[0].BuildUri, out failure);
                        writeLogFunction.Invoke(EventType.Delete, Parent.TeamProjectName, result.Values[0].BuildType, "", result.Values[0].BuildUri, "");
                    }
	            }
	            catch (Exception ex)
	            {
                    writeLogFunction.Invoke(EventType.Delete, Parent.TeamProjectName, result.Values[0].BuildType, "", result.Values[0].BuildUri, "EnforceNumberOfDaysPolicy failed due to: " + ex.ToString());
                }
                finally
                {
                    result.RemoveAt(0);
                }
            }
        }

        private void EnforceNumberOfBuildsPolicy(SortedList<DateTime, TeamBuild> result, DeleteBuildDelegate buildDeleteFunction, WriteLogEntryDelegate writeLogFunction)
        {
            string failure = "";

            while (result.Count > RetentionQuantity)
            {
                try
                {
                    if (OnlyCleanDropLocation)
                    {
                        if (Directory.Exists(result.Values[0].DropLocation))
                        {
                            Directory.Delete(result.Values[0].DropLocation, true);
                            writeLogFunction.Invoke(EventType.Delete, Parent.TeamProjectName, result.Values[0].BuildType, "", result.Values[0].BuildUri, "");
                        }
                    }
                    else
                    {
                        buildDeleteFunction.Invoke(result.Values[0].BuildUri, out failure);
                        writeLogFunction.Invoke(EventType.Delete, Parent.TeamProjectName, result.Values[0].BuildType, "", result.Values[0].BuildUri, "");
                    }
                }
                catch (Exception ex)
                {
                    writeLogFunction.Invoke(EventType.Delete, Parent.TeamProjectName, result.Values[0].BuildType, "", result.Values[0].BuildUri, "EnforceNumberOfBuildsPolicy failed due to: " + ex.ToString());
                }
                finally
                {
                    result.RemoveAt(0);
                }
            }
        }
    }
}
