using System;
using System.Collections.Generic;
using System.Text;
using System.ServiceModel;
using System.Runtime.Serialization;
using System.Diagnostics;
using System.Transactions;
using Callista.BuildLab.Core.Data;
using Callista.BuildLab.Core.Model;

namespace Callista.BuildLab.Core
{
    public interface IEventNotificationCallback
    {
        [OperationContract(IsOneWay = true)]
        void Notify(SubscriptionService.EventType eventType, string eventXml);
    }

    [ServiceContract(Namespace = "http://schemas.callista.se/TFSBuildManager/2007/04/Services/Administration", CallbackContract = typeof(IEventNotificationCallback))]
    public interface IAdministration
    {
        [OperationContract()]
        void Connect();

        [OperationContract()]
        void Disconnect();

        [OperationContract()]
        void ForceRecaching();
    }

    [ServiceContract(Namespace = "http://schemas.callista.se/TFSBuildManager/2007/04/Services/Statistics")]
    public interface IStatistics
    {
        [OperationContract()]
        List<Model.LogEntry> ListEvents(string teamProject, string buildType, string eventType, DateTime startDTM, DateTime endDTM);
    }

    [ServiceContract(Namespace = "http://schemas.callista.se/TFSBuildManager/2007/04/Services/BuildTypeAdministration")]
    public interface IBuildTypeAdministration
    {
        [OperationContract()]
        List<Model.TeamProject> ListTeamProjects();

        [OperationContract()]
        List<Model.BuildType> ListBuildTypes(string teamProjectName);

        [OperationContract()]
        Model.BuildType LoadBuildType(string teamProjectName, string buildTypeName);

        [OperationContract()]
        void CopyBuildType(string sourceTeamProjectName, string sourceBuildTypeName, string targetTeamProjectName, string targetBuildTypeName);
    }

    [ServiceContract(Namespace = "http://schemas.callista.se/TFSBuildManager/2007/04/Services/TriggerAdministration")]
    public interface ITriggerAdministration
    {
        [OperationContract()]
        void AddTriggerToBuildType(string projectName, string buildType, Trigger trigger);

        [OperationContract()]
        void UpdateTrigger(Trigger trigger);

        [OperationContract()]
        void DeleteTrigger(int id);
    }

    [ServiceContract(Namespace = "http://schemas.callista.se/TFSBuildManager/2007/04/Services/RetentionPolicyAdministration")]
    public interface IRetentionPolicyAdministration
    {
        [OperationContract()]
        void AddRetentionPolicyToBuildType(string projectName, string buildType, BuildStatus buildStatus, BuildQuality buildQuality, RetentionPolicyType type, int quantity);

        [OperationContract()]
        void UpdateRetentionPolicy(RetentionPolicy policy);

        [OperationContract()]
        void DeleteRetentionPolicy(int id);
    }

    [ServiceContract(Namespace = "http://schemas.callista.se/TFSBuildManager/2007/04/Services/BuildRepositoryManagement")]
    public interface IBuildRepositoryManagement
    {
        [OperationContract()]
        List<Model.TeamBuild> ListBuilds(string teamProject, string buildType, string buildStatus);

        [OperationContract()]
        Model.TeamBuild GetLatestBuild(string teamProject, string buildType, string buildStatus);

        [OperationContract()]
        Model.TeamBuild GetSpecificBuild(string teamProjectName, string buildNumber);

        [OperationContract()]
        Dictionary<string, string> GetOverriddenParameters(string buildUri);

        [OperationContract()]
        string DeleteBuild(string buildUri);

        [OperationContract()]
        string StopBuild(string buildUri);
    }

    [ServiceContract(Namespace = "http://schemas.callista.se/TFSBuildManager/2007/04/Services/BuildQueueManagement")]
    public interface IBuildQueueManagement
    {
        [OperationContract()]
        List<Model.QueueEntry> ListQueuedBuilds(string teamProject, string buildType);

        [OperationContract()]
        void MoveBuildToBeginingOfQueue(int id);
     
        [OperationContract()]
        void MoveBuildToEndOfQueue(int id);

        [OperationContract()]
        void AddBuildToQueue(string teamProject, string buildType, string requestedBy, Dictionary<string,string> parameters);

        [OperationContract()]
        void RemoveBuildFromQueue(int id);
    }

    public class AdministrationService : IAdministration, 
                                         IStatistics, 
                                         IBuildTypeAdministration, 
                                         ITriggerAdministration, 
                                         IRetentionPolicyAdministration,
                                         IBuildRepositoryManagement,
                                         IBuildQueueManagement
    {
        #region Client Callbacks Implementation

        static List<IEventNotificationCallback> _callbacks = new List<IEventNotificationCallback>();
        public static void CallClients(SubscriptionService.EventType eventType, string eventXml)
        {
            Action<IEventNotificationCallback> invoke = delegate(IEventNotificationCallback callback)
                                                        {
                                                            callback.Notify(eventType, eventXml);
                                                        };

            Trace.TraceInformation(String.Format("{0} AdminstrationService - Started notification of connected clients", DateTime.Now));
            for (int i = 0; i < _callbacks.Count; i++)
            {
                try
                {
                    invoke(_callbacks[i]);
                }
                catch (Exception)
                {
                    _callbacks.RemoveAt(i);
                    i--;
                }
            }
            Trace.TraceInformation(String.Format("{0} AdminstrationService - Finnished notification of connected clients", DateTime.Now));
        }

        #endregion

        #region IAdministration Members

        void IAdministration.Connect()
        {
            try
            {
                IEventNotificationCallback callback = OperationContext.Current.GetCallbackChannel<IEventNotificationCallback>();
                if (!_callbacks.Contains(callback))
                {
                    _callbacks.Add(callback);
                    Trace.TraceInformation(String.Format("{0} AdminstrationService - Client connected (currently {1} clients connecte)", DateTime.Now, _callbacks.Count));
                }
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IAdministration.Connect -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        void IAdministration.Disconnect()
        {
            try
            {
                IEventNotificationCallback callback = OperationContext.Current.GetCallbackChannel<IEventNotificationCallback>();
                if (_callbacks.Contains(callback))
                {
                    _callbacks.Remove(callback);
                    Trace.TraceInformation(String.Format("{0} AdminstrationService - Client disconnected (currently {1} clients connecte)", DateTime.Now, _callbacks.Count));
                }
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IAdministration.Disconnect -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        void IAdministration.ForceRecaching()
        {
            try
            {
                TFS.ResetAllCaches();
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IAdministration.ForceRecaching -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        #endregion

        #region IStatistics Members

        List<LogEntry> IStatistics.ListEvents(string teamProject, string buildType, string eventType, DateTime startDTM, DateTime endDTM)
        {
            try
            {
                IBuildLogRepository db = new BuildLogRepository();
                return db.ListBuildEvents(teamProject, buildType, eventType, startDTM, endDTM);
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IStatistics.ListEvents -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        #endregion

        #region IBuildTypeAdministration Members

        List<Model.TeamProject> IBuildTypeAdministration.ListTeamProjects()
        {
            try
            {
                return TFS.ListTeamProjects();
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IBuildTypeAdministration.ListTeamProjects -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        List<BuildType> IBuildTypeAdministration.ListBuildTypes(string teamProjectName)
        {
            try
            {
                List<BuildType> buildTypeList = new List<BuildType>();

                if (teamProjectName == String.Empty)
                {
                    List<TeamProject> projectList = TFS.ListTeamProjects();
                    foreach (TeamProject p in projectList)
                    {
                        foreach (BuildType bt in p.BuildTypes.Values)
                        {
                            buildTypeList.Add(bt);
                        }
                    }
                }
                else
                {
                    buildTypeList = TFS.BuildService.ListBuildTypes(teamProjectName);
                }

                return buildTypeList;
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IBuildTypeAdministration.ListBuildTypes -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        BuildType IBuildTypeAdministration.LoadBuildType(string teamProjectName, string buildTypeName)
        {
            try
            {
                return TFS.BuildService.LoadBuildType(teamProjectName, buildTypeName);
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IBuildTypeAdministration.LoadBuildType -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        void IBuildTypeAdministration.CopyBuildType(string sourceTeamProjectName, string sourceBuildTypeName, string targetTeamProjectName, string targetBuildTypeName)
        {
            try
            {
                using(TransactionScope scope = new TransactionScope())
                {
                    IBuildTypeRepository db = new BuildTypeRepository();
                    db.CopyBuildType(sourceTeamProjectName, sourceBuildTypeName, targetTeamProjectName, targetBuildTypeName);
                    scope.Complete();
                }

                //Must reset to see update info...
                TFS.ResetAllCaches();
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IBuildTypeAdministration.CopyBuildType -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        #endregion

        #region ITriggerAdministration Members

        void ITriggerAdministration.AddTriggerToBuildType(string projectName, string buildType, Trigger trigger)
        {
            try
            {
                IBuildTypeRepository db = new BuildTypeRepository();
                db.AddTriggerToBuildType(projectName, buildType, trigger);

                //Must reset to see update info...
                TFS.ResetAllCaches();

                //Reschedule the worker thread...
                if (trigger.Type == TriggerType.Schedule)
                {
                    BuildManager.SignalScheduleManagementWorker();
                }
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.ITriggerAdministration.AddTriggerToBuildType -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        void ITriggerAdministration.UpdateTrigger(Trigger trigger)
        {
            try
            {
                IBuildTypeRepository db = new BuildTypeRepository();
                db.UpdateTrigger(trigger);

                //Must reset to see update info...
                TFS.ResetAllCaches();

                //Reschedule the worker thread...
                if (trigger.Type == TriggerType.Schedule)
                {
                    BuildManager.SignalScheduleManagementWorker();
                }
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.ITriggerAdministration.UpdateTrigger -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        void ITriggerAdministration.DeleteTrigger(int id)
        {
            try
            {
                IBuildTypeRepository db = new BuildTypeRepository();
                db.DeleteTrigger(id);

                //Must reset to see update info...
                TFS.ResetAllCaches();

                //Reschedule the worker thread...
                BuildManager.SignalScheduleManagementWorker();
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.ITriggerAdministration.DeleteTrigger -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        #endregion

        #region IRetentionPolicyAdministration Members

        void IRetentionPolicyAdministration.AddRetentionPolicyToBuildType(string projectName, string buildType, BuildStatus buildStatus, BuildQuality buildQuality, RetentionPolicyType type, int quantity)
        {
            try
            {
                IBuildTypeRepository db = new BuildTypeRepository();
                db.AddRetentionPolicyToBuildType(projectName, buildType, buildStatus, buildQuality, type, quantity);

                //Must reset to see update info...
                TFS.ResetAllCaches();
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IRetentionPolicyAdministration.AddRetentionPolicyToBuildType -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        void IRetentionPolicyAdministration.UpdateRetentionPolicy(RetentionPolicy policy)
        {
            try
            {
                IBuildTypeRepository db = new BuildTypeRepository();
                db.UpdateRetentionPolicy(policy);

                //Must reset to see update info...
                TFS.ResetAllCaches();
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IRetentionPolicyAdministration.UpdateRetentionPolicy -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        void IRetentionPolicyAdministration.DeleteRetentionPolicy(int id)
        {
            try
            {
                IBuildTypeRepository db = new BuildTypeRepository();
                db.DeleteRetentionPolicy(id);

                //Must reset to see update info...
                TFS.ResetAllCaches();
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IRetentionPolicyAdministration.DeleteRetentionPolicy -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }
        
        #endregion

        #region IBuildRepositoryManagement Members

        List<Model.TeamBuild> IBuildRepositoryManagement.ListBuilds(string teamProject, string buildType, string buildStatus)
        {
            try
            {
                List<Model.TeamBuild> builds = new List<TeamBuild>();

                if (teamProject == string.Empty)
                {
                    builds = TFS.BuildService.ListBuilds();

                    //If we dont have a project filter but a buildType remove any unwanted projects
                    if (buildType != string.Empty)
                    {
                        List<Model.TeamBuild> temp = new List<TeamBuild>();
                        foreach (TeamBuild tb in builds)
                        {
                            if (tb.BuildType == buildType)
                            {
                                temp.Add(tb);
                            }
                        }
                        builds = temp;
                    }
                }
                else if (buildType == string.Empty)
                {
                    builds = TFS.BuildService.ListBuilds(teamProject);
                }
                else
                {
                    builds = TFS.BuildService.ListBuilds(teamProject, buildType);
                }

                //If we have a status filter remove any unwanted builds
                if (buildStatus != string.Empty)
                {
                    List<TeamBuild> temp = new List<TeamBuild>();
                    foreach (TeamBuild tb in builds)
                    {
                        if (tb.Status == (BuildStatus)Enum.Parse(typeof(BuildStatus), buildStatus))
                        {
                            temp.Add(tb);
                        }
                    }
                    builds = temp;
                }

                return builds;
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IBuildRepositoryManagement.ListBuilds -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        Model.TeamBuild IBuildRepositoryManagement.GetLatestBuild(string teamProject, string buildType, string buildStatus)
        {
            try
            {
                Model.TeamBuild latestBuild = null;
                IBuildRepositoryManagement buildRepository = this;

                List<Model.TeamBuild> builds = buildRepository.ListBuilds(teamProject, buildType, buildStatus);
                foreach (Model.TeamBuild build in builds)
                {
                    if (latestBuild == null)
                    {
                        latestBuild = build;
                    }
                    else
                    {
                        if (build.FinishTime > latestBuild.FinishTime)
                        {
                            latestBuild = build;
                        }
                    }
                }

                return latestBuild;
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IBuildRepositoryManagement.GetLatestBuild -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        TeamBuild IBuildRepositoryManagement.GetSpecificBuild(string teamProjectName, string buildNumber)
        {
            try
            {
                return TFS.BuildService.LoadBuild(teamProjectName, buildNumber);
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IBuildRepositoryManagement.GetSpecificBuild -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        Dictionary<string, string> IBuildRepositoryManagement.GetOverriddenParameters(string buildUri)
        {
            try
            {
                QueueEntry qe = BuildManager.QueryRegisteredBuilds(buildUri);

                if (qe != null)
                {
                    return qe.OverriddenBuildParameters;
                }

                return null;
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IBuildRepositoryManagement.GetOverriddenParameters -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        string IBuildRepositoryManagement.DeleteBuild(string buildUri)
        {
            try
            {
                string message = "";
                TFS.BuildService.DeleteBuild(buildUri, out message);
                return message;
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IBuildRepositoryManagement.DeleteBuild -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        string IBuildRepositoryManagement.StopBuild(string buildUri)
        {
            try
            {
                string message = "";
                TFS.BuildService.StopBuild(buildUri, out message);
                return message;
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IBuildRepositoryManagement.StopBuild -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        #endregion

        #region IBuildQueueManagement Members

        List<Model.QueueEntry> IBuildQueueManagement.ListQueuedBuilds(string teamProject, string buildType)
        {
            try
            {
                IBuildQueueRepositoryAdmin admin = new BuildQueueRepository();
                List<Model.QueueEntry> list = admin.ListQueuedBuilds(teamProject, buildType);
                return list;
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IBuildQueueManagement.ListQueuedBuilds -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        void IBuildQueueManagement.MoveBuildToBeginingOfQueue(int id)
        {
            try
            {
                IBuildQueueRepositoryAdmin admin = new BuildQueueRepository();
                admin.MoveBuildToBeginingOfQueue(id);
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IBuildQueueManagement.MoveBuildToBeginingOfQueue -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        void IBuildQueueManagement.MoveBuildToEndOfQueue(int id)
        {
            try
            {
                IBuildQueueRepositoryAdmin admin = new BuildQueueRepository();
                admin.MoveBuildToEndOfQueue(id);
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IBuildQueueManagement.MoveBuildToEndOfQueue -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        void IBuildQueueManagement.RemoveBuildFromQueue(int id)
        {
            try
            {
                IBuildQueueRepositoryAdmin admin = new BuildQueueRepository();
                admin.RemoveBuildFromQueue(id);
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IBuildQueueManagement.RemoveBuildFromQueue -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        void IBuildQueueManagement.AddBuildToQueue(string teamProject, string buildType, string requestedBy, Dictionary<string, string> parameters)
        {
            try
            {
                IBuildQueueRepository db = new BuildQueueRepository();
                BuildType bt = TFS.BuildService.LoadBuildType(teamProject, buildType);
                db.PushBuildToQueue(new QueueEntry(teamProject, buildType, bt.DefaultBuildServer, requestedBy, parameters));
            }
            catch (Exception ex)
            {
                Trace.TraceError(String.Format("{0} AdminstrationService.IBuildQueueManagement.AddBuildToQueue -- {1}", DateTime.Now, ex.ToString()));
                throw new FaultException(ex.Message);
            }
        }

        #endregion
    }
}
