using System;
using System.Text;
using System.Reflection;
using System.Threading;
using System.Diagnostics;
using System.ServiceModel;
using System.Configuration;
using System.Collections.Generic;
using Callista.BuildLab.Core.Data;
using Callista.BuildLab.Core.Model;

namespace Callista.BuildLab.Core
{
    public class BuildManager
    {
        public delegate void TriggerQueueDelegate();

        private class MonitoredBuild
        {
            public string _uri;
            public QueueEntry _qe;
            public TriggerQueueDelegate _trigger;

            public MonitoredBuild(string uri, QueueEntry qe, TriggerQueueDelegate trigger)
            {
                _uri = uri;
                _qe = qe;
                _trigger = trigger;
            }
        }
        
        private Thread _queueWorkerThread = null;
        private int _queueManagementTimeout = 60000;

        private Thread _scheduleWorkerThread = null;
        private int _scheduleManagementTimeout = 60000;

        private Thread _retentionEnforcementWorkerThread = null;
        private int _retentionEnforcementTimeout = 3600000;
        
        private bool _isPausing = false; 
        private bool _isStopping = false;

        private static ServiceHost _notificationHost = null;
        private static ServiceHost _administrationHost = null;
        private static Dictionary<string, MonitoredBuild> _monitoredBuild = new Dictionary<string, MonitoredBuild>();

        private static AutoResetEvent _queueManagementEvent = new AutoResetEvent(false);
        private static AutoResetEvent _scheduleManagementEvent = new AutoResetEvent(false);
        private static AutoResetEvent _retentionEnforcementEvent = new AutoResetEvent(false);

        public static void SignalQueueManagementWorker()
        {
            lock (_queueManagementEvent)
            {
                _queueManagementEvent.Set();
            }
        }

        public static void SignalScheduleManagementWorker()
        {
            lock (_scheduleManagementEvent)
            {
                _scheduleManagementEvent.Set();
            }
        }

        public static void SignalRetentionEnforcementWorker()
        {
            lock (_retentionEnforcementEvent)
            {
                _retentionEnforcementEvent.Set();
            }
        }

        public static ServiceHost NotificationHost
        {
            get { return _notificationHost; }
        }

        public static ServiceHost AdministrationHost
        {
            get { return _administrationHost; }
        }

        public static void RegisterBuildForNotification(string uri, QueueEntry qe, TriggerQueueDelegate triggerQueueFunction)
        {
            lock (_monitoredBuild)
            {
                if (!uri.StartsWith(@"vstfs:///Build/Build/"))
                {
                    uri = @"vstfs:///Build/Build/" + uri;
                }

                if (_monitoredBuild.ContainsKey(uri))
                {
                    MonitoredBuild mb = _monitoredBuild[uri];
                    if (mb._trigger == null)
                    {
                        mb._trigger = triggerQueueFunction;
                    }
                }
                else
                {
                    Trace.TraceInformation(String.Format("{0} BuildManager - Registered build notification for build uri {1}", DateTime.Now, uri));
                    _monitoredBuild.Add(uri, new MonitoredBuild(uri, qe, triggerQueueFunction));
                }
            }
        }

        public static QueueEntry UnregisterBuildFromNotification(string uri)
        {
            lock (_monitoredBuild)
            {
                if (!uri.StartsWith(@"vstfs:///Build/Build/"))
                {
                    uri = @"vstfs:///Build/Build/" + uri;
                }

                if (_monitoredBuild.ContainsKey(uri))
                {
                    MonitoredBuild mb = _monitoredBuild[uri];
                    if (mb._trigger != null)
                    {
                        Trace.TraceInformation(String.Format("{0} BuildManager - Invoking callback for registered build notification for build uri {1}", DateTime.Now, uri));
                        mb._trigger.Invoke();
                    }
                    Trace.TraceInformation(String.Format("{0} BuildManager - Unregistered build notification for build uri {1}", DateTime.Now, uri));
                    _monitoredBuild.Remove(uri);
                    return mb._qe;
                }
            }
            return null;
        }

        public static QueueEntry QueryRegisteredBuilds(string uri)
        {
            lock (_monitoredBuild)
            {
                if (!uri.StartsWith(@"vstfs:///Build/Build/"))
                {
                    uri = @"vstfs:///Build/Build/" + uri;
                }

                if (_monitoredBuild.ContainsKey(uri))
                {
                    Trace.TraceInformation(String.Format("{0} BuildManager - Querying queue entry information for monitored build uri {1}", DateTime.Now, uri));
                    return _monitoredBuild[uri]._qe;
                }
            }
            return null;
        }

        public bool IsPausing
        {
            get { return _isPausing; }
            set { lock (this) { _isPausing = value; } }
        }

        public bool IsStopping
        {
            get { return _isStopping; }
            set { lock (this) { _isStopping = value; } }
        }

        private bool IsServiceActive()
        {
            while (IsPausing && !IsStopping)
                Thread.Sleep(1000);

            if (IsStopping)
                return false;

            return true;
        }

        private void WriteToEventLog(string message)
        {
            EventLog.WriteEntry("TfsBuildlab", String.Format("{0} {1}", DateTime.Now, message), EventLogEntryType.Information);
        }

        private void WriteToEventLog(string message, Exception ex)
        {
            Trace.TraceError(String.Format("{0} {1}: An exception occured ... {2}", DateTime.Now, message, ex.ToString()));
            EventLog.WriteEntry("TfsBuildlab", String.Format("{0} {1}: An exception occured ... {2}", DateTime.Now, message, ex.ToString()), EventLogEntryType.Error);
        }
	
        private void Initialize()
        {
            try
            {
                Trace.TraceInformation(String.Format("{0} BuildManager initialization started", DateTime.Now));

                // Start administration service
                OpenAdministrationService();

                // Start notification service
                OpenNotificationService();

                // Setup event registration (no bissubscribe)
                UpdateSubscriptionService();
                
                _queueManagementTimeout = Int32.Parse(ConfigurationManager.AppSettings["QueueManagementInterval"]);
                _scheduleManagementTimeout = Int32.Parse(ConfigurationManager.AppSettings["ScheduleManagementInterval"]);
                _retentionEnforcementTimeout = Int32.Parse(ConfigurationManager.AppSettings["RetentionPolicyEnforcementInterval"]);

                _queueManagementEvent = new AutoResetEvent(false);
                _scheduleManagementEvent = new AutoResetEvent(false);
                _retentionEnforcementEvent = new AutoResetEvent(false);

                _queueWorkerThread = new Thread(this.HandleQueue);
                _queueWorkerThread.Start();

                _scheduleWorkerThread = new Thread(this.HandleSchedules);
                _scheduleWorkerThread.Start();

                _retentionEnforcementWorkerThread = new Thread(this.HandleCleanup);
                _retentionEnforcementWorkerThread.Start();

                WriteToEventLog("BuildManager has initialized");

                Trace.TraceInformation(String.Format("{0} BuildManager initialization Completed", DateTime.Now));
            }
            catch (Exception ex)
            {
                WriteToEventLog("BuildManager failed to initialize", ex);
            }
        }

        private void Terminate()
        {
            _isStopping = true;
            CloseAdministratonService();
            CloseNotificationService();
        }

        public void Start()
        {
            Initialize();
            Trace.TraceInformation(String.Format("{0} BuildManager started.", DateTime.Now));
        }

        public void Stop()
        {
            Terminate();
            Trace.TraceInformation(String.Format("{0} BuildManager stopped.", DateTime.Now));
        }

        public void Pause()
        {
            IsPausing = true;

            CloseNotificationService();

            Trace.TraceInformation(String.Format("{0} BuildManager paused.", DateTime.Now));
        }

        public void Resume()
        {
            IsPausing = false;

            OpenNotificationService();

            Trace.TraceInformation(String.Format("{0} BuildManager resumed.", DateTime.Now));
        }

        private void HandleSchedules()
        {
            while (IsServiceActive())
            {
                try
                {
                    SortedList<string, List<BuildType>> buildTypes = new SortedList<string, List<BuildType>>();

                    List<BuildType> nextToRunBT = null;
                    TimeSpan nextToRunTS = TimeSpan.MinValue;

                    //Refresh list of buildtypes to schedule
                    foreach (TeamProject tp in TFS.ListTeamProjects())
                    {
                        foreach (BuildType bt in tp.BuildTypes.Values)
                        {
                            foreach (Trigger t in bt.Triggers.Values)
                            {
                                if (t.Type == TriggerType.Schedule)
                                {
                                    if (!buildTypes.ContainsKey(t.Data))
                                    {
                                        buildTypes.Add(t.Data, new List<BuildType>());
                                    }
                                    buildTypes[t.Data].Add(bt);
                                }
                            }
                        }
                    }
                    
                    // Find next schedule to run today.
                    foreach (string start in buildTypes.Keys)
                    {
                        TimeSpan when = new TimeSpan(Int32.Parse(start.Substring(0, 2)),Int32.Parse(start.Substring(2, 2)),0);
                        TimeSpan now = new TimeSpan(DateTime.Now.Hour,DateTime.Now.Minute,0);

                        if (when > now)
                        {
                            nextToRunBT = buildTypes[start];
                            nextToRunTS = when - now;
                            Trace.TraceInformation(String.Format("{0} BuildManager - Build Scheduling found to schedule today [TeamProject={1}BuildType={2},When={3}]", DateTime.Now, nextToRunBT[0].TeamProjectName, nextToRunBT[0].Name, nextToRunTS.ToString()));
                            break;
                        }
                    }

                    // Nothing found to schedule today, schedule the earliest for tomorrow.
                    if (nextToRunBT == null && buildTypes.Count > 0)
                    {
                        TimeSpan when = new TimeSpan(1, Int32.Parse(buildTypes.Keys[0].Substring(0, 2)), Int32.Parse(buildTypes.Keys[0].Substring(2, 2)), 0);
                        TimeSpan now = new TimeSpan(DateTime.Now.Hour, DateTime.Now.Minute, 0);

                        if (when > now)
                        {
                            nextToRunBT = buildTypes.Values[0];
                            nextToRunTS = when - now;
                            Trace.TraceInformation(String.Format("{0} BuildManager - Build Scheduling nothing found to schedule today, schedule the earliest for tomorrow [TeamProject={1}BuildType={2},When={3}]", DateTime.Now, nextToRunBT[0].TeamProjectName, nextToRunBT[0].Name, nextToRunTS.ToString()));
                        }
                    }

                    if (IsServiceActive())
                    {
                        if (nextToRunBT != null)
                        {
                            //if timeout occurs we have waited the aloted time and should execute otherwise 
                            //we have a new schedule to consider and should reschedule...

                            if (!_scheduleManagementEvent.WaitOne(nextToRunTS, false))
                            {
                                string teamProject = "";
                                string buildType = "";

                                try
                                {
                                    IBuildQueueRepository db = new BuildQueueRepository();

                                    foreach (BuildType bt in nextToRunBT)
                                    {
                                        teamProject = bt.TeamProjectName;
                                        buildType = bt.Name;

                                        db.PushBuildToQueue(new QueueEntry(bt.TeamProjectName, bt.Name, bt.DefaultBuildServer));

                                        Trace.TraceInformation(String.Format("{0} BuildManager - Scheduled build put on queue [TeamProject={1}BuildType={2}]", DateTime.Now, bt.TeamProjectName, bt.Name));
                                    }
                                }
                                catch (Exception ex)
                                {
                                    BuildLog.WriteEntryToLog(EventType.Error, teamProject, buildType, "", "", ex.ToString());
                                }

                                nextToRunTS = TimeSpan.MinValue;
                                nextToRunBT = null;
                            }
                        }
                    }

                    _scheduleManagementEvent.WaitOne(_scheduleManagementTimeout, false);
                }
                catch (Exception ex)
                {
                    WriteToEventLog("BuildManager.HandleSchedules", ex);
                }
            }
        }

        private void HandleCleanup()
        {
            while (IsServiceActive())
            {
                try
                {
                    foreach (TeamProject tp in TFS.ListTeamProjects())
                    {
                        foreach (BuildType bt in tp.BuildTypes.Values)
                        {
                            foreach (RetentionPolicy rp in bt.RetentionPolicies.Values)
                            {
                                // Refresh build list for each individual retention policy
                                List<TeamBuild> builds = TFS.BuildService.ListBuilds(tp.Name, bt.Name);

                                if (IsServiceActive()) 
                                {
                                    Trace.TraceInformation(String.Format("{0} BuildManager - Begining to enforce retention policies for {1} - {2}", DateTime.Now, tp.Name, bt.Name));
                                    rp.EnforceRetentionPolicy(builds, TFS.BuildService.DeleteBuild, BuildLog.WriteEntryToLog);
                                    Trace.TraceInformation(String.Format("{0} BuildManager - Done enforcing of retention policies for {1} - {2}", DateTime.Now, tp.Name, bt.Name));
                                }
                                else 
                                {
                                    return; // Stop processing since service is shutting down.
                                }
                            }
                        }
                    }

                    _retentionEnforcementEvent.WaitOne(_retentionEnforcementTimeout, false);
                }
                catch (Exception ex)
                {
                    WriteToEventLog("BuildManager.HandleCleanup", ex);
                }
            }
        }

        private void HandleQueue()
        {
            while (IsServiceActive())
            {
                try
                {
                    IBuildQueueRepository db = new BuildQueueRepository();
                    
                    long count = db.QueueEntryCount();
                    long current = -1;

                    for(long index=0;index<count;index++)
                    {
                        string buildUri = "";
                        
                        BuildType buildType = null;
                        List<string> servers = null;

                        if (IsServiceActive())
                        {
                            QueueEntry nextEntry = db.PeekNextBuildOnQueue(current);
                            if (nextEntry != null)
                            {
                                current = nextEntry.ID;

                                buildType = TFS.BuildService.LoadBuildType(nextEntry.TeamProjectName, nextEntry.BuildTypeName);
                                if (nextEntry.TriggeringChangesetID.HasValue)
                                {
                                    nextEntry.InvolvedChangesets = TFS.BuildService.GetLatestChangesetsForBuildType(buildType, nextEntry.TriggeringChangesetID.Value);
                                }
                                servers = buildType.GetAvailableBuildServers(nextEntry.InvolvedChangesets);

                                foreach (string s in servers)
                                {
                                    try
                                    {
                                        buildUri = TFS.BuildService.StartBuild(buildType, s);
                                        if (!string.IsNullOrEmpty(buildUri))
                                        {
                                            break;
                                        }
                                    }
                                    catch (Exception ex)
                                    {
                                        if (!ex.Message.Contains("TF42045: A build for the team project:"))
                                        {
                                            BuildLog.WriteEntryToLog(EventType.Error, nextEntry.TeamProjectName, nextEntry.BuildTypeName, "", "", ex.ToString());
                                            Trace.TraceError(String.Format("{0} Unable to started build from queue an exception occured {1}", DateTime.Now, ex.ToString()));
                                            db.IncrementBuildDeliveryAttempts(current);
                                        }
                                    }
                                }
                            
                                if (!string.IsNullOrEmpty(buildUri))
                                {
                                    BuildManager.RegisterBuildForNotification(buildUri, nextEntry, this.TriggerQueue);
                                    db.PopBuildFromQueue(current);

                                    if (nextEntry.TriggeringChangesetID.HasValue)
                                    {
                                        BuildLog.WriteEntryToLog(EventType.StartCheckin, buildType.TeamProjectName, buildType.Name, nextEntry.TriggeringUser, buildUri);
                                        Trace.TraceInformation(String.Format("{0} BuildManager - Started build from queue of type checkin trigger for {1} - {2} with the resulting build uri {3}", DateTime.Now, buildType.TeamProjectName, buildType.Name, buildUri));
                                    }
                                    else
                                    {
                                        BuildLog.WriteEntryToLog(EventType.StartScheduled, buildType.TeamProjectName, buildType.Name, "", buildUri);
                                        Trace.TraceInformation(String.Format("{0} BuildManager - Started build from queue of type schedule trigger for {1} - {2} with the resulting build uri {3}", DateTime.Now, buildType.TeamProjectName, buildType.Name, buildUri));
                                    }
                                }
                            }
                        }
                        else 
                        {
                            return; // Stop processing since service is shutting down.
                        }
                    }

                    _queueManagementEvent.WaitOne(_queueManagementTimeout, false);
                }
                catch (Exception ex)
                {
                    WriteToEventLog("BuildManager.HandleQueue", ex);
                }
            }
        }

        private static void UpdateSubscriptionService()
        {
            string autoRegisterNotifications = ConfigurationManager.AppSettings["AutoRegisterNotifications"];

            if (string.IsNullOrEmpty(autoRegisterNotifications))
            {
                autoRegisterNotifications = "true";
            }

            if (!bool.Parse(autoRegisterNotifications))
            {
                Trace.TraceWarning(String.Format("{0} BuildManager - Auto registration of notifications is turned off make sure you have the proper subscriptions registered.", DateTime.Now));
                return;
            }

            Trace.TraceInformation(String.Format("{0} BuildManager - UpdateSubscriptionService started updating of event subscriptions.", DateTime.Now));
            TFS.SubscriptionService.UpdateSubscriptions();
            Trace.TraceInformation(String.Format("{0} BuildManager - UpdateSubscriptionService finnished updating of event subscriptions.", DateTime.Now));
        }

        private void OpenNotificationService()
        {
            string path = string.Format("http://{0}:{1}", 
                                        Environment.MachineName,
                                        ConfigurationManager.AppSettings["buildLabPort"]);
            
            BasicHttpBinding binding = new BasicHttpBinding();

            binding.ReaderQuotas.MaxArrayLength = int.MaxValue;
            binding.ReaderQuotas.MaxBytesPerRead = int.MaxValue;
            binding.ReaderQuotas.MaxDepth = int.MaxValue;
            binding.ReaderQuotas.MaxStringContentLength = int.MaxValue;
            binding.ReaderQuotas.MaxBytesPerRead = int.MaxValue;

            binding.MaxReceivedMessageSize = int.MaxValue;
            binding.MaxBufferSize = int.MaxValue;

            _notificationHost = new ServiceHost(typeof(NotificationService), new Uri(path));
            _notificationHost.AddServiceEndpoint(typeof(INotificationService), binding, "");
            _notificationHost.Open();

            Trace.TraceInformation(String.Format("{0} BuildManager - NotificationService started.", DateTime.Now));
        }

        private void CloseNotificationService()
        {
            if (_notificationHost != null)
            {
                if (_notificationHost.State == CommunicationState.Opened)
                {
                    _notificationHost.Close();
                }
                _notificationHost = null;

                Trace.TraceInformation(String.Format("{0} BuildManager - NotificationService stopped.", DateTime.Now));
            }
        }

        private void OpenAdministrationService()
        {
            string path = string.Format("net.tcp://{0}:{1}",
                                        Environment.MachineName,
                                        ConfigurationManager.AppSettings["buildLabAdminPort"]);

            NetTcpBinding binding = new NetTcpBinding(SecurityMode.None);

            binding.ReaderQuotas.MaxArrayLength = int.MaxValue;
            binding.ReaderQuotas.MaxBytesPerRead = int.MaxValue;
            binding.ReaderQuotas.MaxDepth = int.MaxValue;
            binding.ReaderQuotas.MaxStringContentLength = int.MaxValue;
            binding.ReaderQuotas.MaxBytesPerRead = int.MaxValue;

            binding.MaxReceivedMessageSize = int.MaxValue;
            binding.MaxBufferSize = int.MaxValue;

            binding.ReliableSession.Enabled = false;

            binding.OpenTimeout = TimeSpan.MaxValue;
            binding.CloseTimeout = TimeSpan.MaxValue;
            binding.ReceiveTimeout = TimeSpan.MaxValue;
            binding.SendTimeout = TimeSpan.MaxValue;
            
            _administrationHost = new ServiceHost(typeof(AdministrationService), new Uri(path));
            _administrationHost.AddServiceEndpoint(typeof(IAdministration), binding, "");
            _administrationHost.AddServiceEndpoint(typeof(IStatistics), binding, "");
            _administrationHost.AddServiceEndpoint(typeof(IBuildTypeAdministration), binding, "");
            _administrationHost.AddServiceEndpoint(typeof(ITriggerAdministration), binding, "");
            _administrationHost.AddServiceEndpoint(typeof(IRetentionPolicyAdministration), binding, "");
            _administrationHost.AddServiceEndpoint(typeof(IBuildRepositoryManagement), binding, "");
            _administrationHost.AddServiceEndpoint(typeof(IBuildQueueManagement), binding, "");
            _administrationHost.Open();

            Trace.TraceInformation(String.Format("{0} BuildManager - AdministrationService started.", DateTime.Now));
        }

        private void CloseAdministratonService()
        {
            if (_administrationHost != null)
            {
                if (_administrationHost.State == CommunicationState.Opened)
                {
                    _administrationHost.Close();
                }

                Trace.TraceInformation(String.Format("{0} BuildManager - AdministrationService stopped.", DateTime.Now));
            }
        }

        public void TriggerQueue()
        {
            Trace.TraceInformation(String.Format("{0} BuildManager - Registered build notification triggered", DateTime.Now));
        }
    }
}
