using System;
using System.IO;
using System.Xml;
using System.Xml.Serialization;
using System.Text;
using System.Diagnostics;
using System.Collections.Generic;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.Build;
using Microsoft.TeamFoundation.Build.Proxy;
using Microsoft.TeamFoundation.Server;
using Microsoft.TeamFoundation.VersionControl.Client;
using Microsoft.TeamFoundation.VersionControl.Common;
using Callista.BuildLab.Core.Data;
using Callista.BuildLab.Core.Model;

namespace Callista.BuildLab.Core
{
    public class BuildService
    {
        private TeamFoundationServer _tfs;
        private BuildStore _buildStore;
        private BuildController _buildController;
        private VersionControlServer _versionControlServer;

        public BuildService(TeamFoundationServer tfs)
        {
            _tfs = tfs;
            _buildStore = (BuildStore)tfs.GetService(typeof(BuildStore));
            _buildController = (BuildController)tfs.GetService(typeof(BuildController));
            _versionControlServer = (VersionControlServer)_tfs.GetService(typeof(VersionControlServer));
        }

        public void ResetAllCaches()
        {
            _cachedBuildTypes.Clear();
            _cachedBuildTypesForTeamProjects.Clear();
        }

        private void ExtractInformationFromTeamBuildProjectFiles(Item bti, BuildType bt)
        {
            ItemSet buildTypeContents = _versionControlServer.GetItems(bti.ServerItem, VersionSpec.Latest, RecursionType.OneLevel, DeletedState.NonDeleted, ItemType.File);

            for (int i = 0; i < buildTypeContents.Items.Length; i++)
            {
                if (buildTypeContents.Items[i].ServerItem.EndsWith("TFSBuild.proj"))
                {
                    string tempFile = string.Format(@"{0}\TfsBuildLab\{1}_{2}",
                                            Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData),
                                            Guid.NewGuid().ToString(),
                                            VersionControlPath.GetFileName(buildTypeContents.Items[i].ServerItem).Replace(@"/", @"_"));

                    XmlDocument projectFileXml = new XmlDocument();
                    XmlNamespaceManager nsm = new XmlNamespaceManager(projectFileXml.NameTable);

                    nsm.AddNamespace(@"msbuild", @"http://schemas.microsoft.com/developer/msbuild/2003");

                    buildTypeContents.Items[i].DownloadFile(tempFile);
                    projectFileXml.Load(tempFile);

                    bt.DefaultBuildServer = projectFileXml.DocumentElement.SelectSingleNode(@"descendant::msbuild:ProjectExtensions/msbuild:BuildMachine", nsm).InnerText;
                    bt.DefaultBuildDirectory = projectFileXml.DocumentElement.SelectSingleNode(@"descendant::msbuild:PropertyGroup/msbuild:BuildDirectoryPath", nsm).InnerText;

                    XmlNodeList properties = projectFileXml.DocumentElement.SelectNodes(@"descendant::msbuild:PropertyGroup/*", nsm);
                    foreach (XmlNode property in properties)
                    {

                        if (!bt.Parameters.Exists( delegate(BuildParameter bp) { return bp.Name == property.Name; } ))
                        {
                            bt.Parameters.Add(new BuildParameter(property.Name, property.InnerText));
                        }
                        else
                        {
                            Trace.TraceInformation("{0} BuildService - Found duplicate property [{1}] in build script [{2}]", DateTime.Now, property.Name, tempFile);
                        }
                    }
                    
                    File.Delete(tempFile);
                }
                else if(buildTypeContents.Items[i].ServerItem.EndsWith("WorkspaceMapping.xml"))
                {
                    string tempFile = string.Format(@"{0}\TfsBuildLab\{1}_{2}",
                                            Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData),
                                            Guid.NewGuid().ToString(),
                                            VersionControlPath.GetFileName(buildTypeContents.Items[i].ServerItem).Replace(@"/", @"_"));

                    _versionControlServer.DownloadFile(buildTypeContents.Items[i].ServerItem, tempFile);
                    
                    string fileContent = File.ReadAllText(tempFile);
                    SerializedWorkspace sw = Common.Deserialize<SerializedWorkspace>(fileContent);
                    foreach(InternalMapping im in sw.Mappings)
                    {
                        if(im.Type == "Map")
                        {
                            if (!bt.WorkspaceMapping.MappedPaths.Contains(im.ServerItem))
                            {
                                bt.WorkspaceMapping.MappedPaths.Add(im.ServerItem);
                            }
                            else
                            {
                                Trace.TraceInformation("{0} BuildService - Found duplicate map [{1}] in workspacemappings [{2}]", DateTime.Now, im.ServerItem, tempFile);
                            }
                        }
                        else
                        {
                            if (!bt.WorkspaceMapping.CloakedPaths.Contains(im.ServerItem))
                            {
                                bt.WorkspaceMapping.CloakedPaths.Add(im.ServerItem);
                            }
                            else
                            {
                                Trace.TraceInformation("{0} BuildService - Found duplicate cloak [{1}] in workspacemappings [{2}]", DateTime.Now, im.ServerItem, tempFile);
                            }
                        }
                    }

                    File.Delete(tempFile);
                }
            }
        }

        static Dictionary<string, BuildType> _cachedBuildTypes = new Dictionary<string, BuildType>();
        public BuildType LoadBuildType(string teamProjectName, string buildTypeName)
        {
            lock (_cachedBuildTypes)
            {
                if (!_cachedBuildTypes.ContainsKey(teamProjectName + buildTypeName))
                {
                    IBuildTypeRepository db = new BuildTypeRepository();
                    Model.TeamProject project = db.GetTeamProject(teamProjectName);

                    BuildType buildType = null;
                    if (project.BuildTypes.ContainsKey(buildTypeName))
                    {
                        buildType = project.BuildTypes[buildTypeName];
                    }
                    else
                    {
                        buildType = new BuildType(project.Name, buildTypeName);
                    }

                    string path = string.Format("$/{0}/TeamBuildTypes/{1}", teamProjectName, buildTypeName);
                    ExtractInformationFromTeamBuildProjectFiles(_versionControlServer.GetItem(path), buildType);

                    _cachedBuildTypes.Add(teamProjectName + buildTypeName, buildType);
                }

                return _cachedBuildTypes[teamProjectName + buildTypeName];
            }
        }

        static Dictionary<string, List<BuildType>> _cachedBuildTypesForTeamProjects = new Dictionary<string, List<BuildType>>();
        public List<BuildType> ListBuildTypes(string teamProjectName)
        {
            lock (_cachedBuildTypes)
            {
                if (!_cachedBuildTypesForTeamProjects.ContainsKey(teamProjectName))
                {
                    IBuildTypeRepository db = new BuildTypeRepository();
                    Model.TeamProject project = db.GetTeamProject(teamProjectName);
                    List<BuildType> buildTypeList = new List<BuildType>();

                    string path = string.Format("$/{0}/TeamBuildTypes", teamProjectName);
                    ItemSet buildTypeItems = _versionControlServer.GetItems(path, VersionSpec.Latest, RecursionType.OneLevel, DeletedState.NonDeleted, ItemType.Folder);
                    for (int i = 1; i < buildTypeItems.Items.Length; i++)
                    {
                        BuildType buildType = null;
                        string buildTypeName = VersionControlPath.GetFileName(buildTypeItems.Items[i].ServerItem);

                        if (project.BuildTypes.ContainsKey(buildTypeName))
                        {
                            buildType = project.BuildTypes[buildTypeName];
                        }
                        else
                        {
                            buildType = new BuildType(project.Name, buildTypeName);
                        }

                        ExtractInformationFromTeamBuildProjectFiles(buildTypeItems.Items[i], buildType);
                        buildTypeList.Add(buildType);
                    }

                    _cachedBuildTypesForTeamProjects.Add(teamProjectName, buildTypeList);
                }

                return _cachedBuildTypesForTeamProjects[teamProjectName];
            }
        }

        public TeamBuild LoadBuild(string teamProjectName,string buildNumber)
        {
            string uri = _buildStore.GetBuildUri(teamProjectName, buildNumber);
            BuildData bd = _buildStore.GetBuildDetails(uri);
            if (bd != null)
            {
                return new TeamBuild(bd);
            }

            return null;
        }

        public List<TeamBuild> ListBuilds()
        {
            ICommonStructureService css = (ICommonStructureService)_tfs.GetService(typeof(ICommonStructureService));
            List<TeamBuild> builds = new List<TeamBuild>();

            foreach (ProjectInfo pi in css.ListAllProjects())
            {
                builds.AddRange(ListBuilds(pi.Name));
            }
            
            return builds;
        }

        public List<TeamBuild> ListBuilds(string teamProject)
        {
            return ListBuilds(teamProject, null);
        }

        public List<TeamBuild> ListBuilds(string teamProject, string buildType)
        {
            List<TeamBuild> buildList = new List<TeamBuild>();
            foreach (BuildData bd in _buildStore.GetListOfBuilds(teamProject, buildType))
            {
                buildList.Add(new TeamBuild(bd));
            }
            return buildList;
        }

        public string StartBuild(BuildType bt)
        {
            return StartBuild(bt.TeamProjectName, bt.Name, bt.DefaultBuildDirectory, bt.DefaultBuildServer);
        }

        public string StartBuild(BuildType bt, string buildMachine)
        {
            return StartBuild(bt.TeamProjectName, bt.Name, bt.DefaultBuildDirectory, buildMachine);
        }

        public string StartBuild(string teamProject, string buildType, string buildDirectory, string buildMachine)
        {
            BuildParameters bp = new BuildParameters();
            bp.TeamFoundationServer = _tfs.Uri.AbsoluteUri; // "tfs";
            bp.TeamProject = teamProject;   // "myproj"
            bp.BuildType = buildType;   // ContinuousIntegration
            bp.BuildDirectory = buildDirectory; // @"C:\Source\";
            bp.BuildMachine = buildMachine; // "tfs";

            return _buildController.StartBuild(bp);
        }

        public bool StopBuild(string buildUri, out string result)
        {
            return _buildController.StopBuild(buildUri, out result);
        }

        public bool DeleteBuild(string buildUri, out string failureMessage)
        {
            return _buildController.DeleteBuild(buildUri, out failureMessage);
        }

        public List<BuildType> GetAffectedBuildTypes(Core.Model.Changeset cs)
        {
            List<BuildType> buildTypeList = new List<BuildType>();

            foreach (Model.TeamProject teamProject in TFS.ListTeamProjects())
            {
                foreach (BuildType bt in teamProject.BuildTypes.Values)
                {
                    if(bt.WorkspaceMapping.IsMatching(cs) && !buildTypeList.Contains(bt))
                    {
                        buildTypeList.Add(bt);
                    }
                }
            }

            return buildTypeList;
        }

        public List<Core.Model.Changeset> GetLatestChangesetsForBuildType(BuildType bt,int triggeringChangesetID)
        {
            List<Core.Model.Changeset> changesetList = new List<Core.Model.Changeset>();

            foreach (Microsoft.TeamFoundation.VersionControl.Client.Changeset cs in _versionControlServer.QueryHistory(bt.WorkspaceMapping.MappedPaths[0], VersionSpec.Latest, 0, RecursionType.Full, null, VersionSpec.ParseSingleSpec(triggeringChangesetID.ToString(), null), null, Int32.MaxValue, true, false))
            {
                changesetList.Add(Core.Model.Changeset.CreateFromObject(cs));
            }

            return changesetList;
        }
    }
}
