using System;
using System.Net.Mail;
using System.Collections.Generic;
using System.Configuration;
using System.DirectoryServices;
using System.Text;
using System.ServiceModel;
using System.Runtime.Serialization;
using System.Xml;
using System.Xml.Serialization;
using System.IO;
using System.Diagnostics;
using Callista.BuildLab.Core.Data;
using Callista.BuildLab.Core.Model;
using Microsoft.TeamFoundation.Build.Proxy;

namespace Callista.BuildLab.Core
{
    [DataContract(Namespace = "http://schemas.microsoft.com/TeamFoundation/2005/06/Services/Notification/03")]
    public class SubscriptionInfo
    {
        [DataMember()]
        public string ID;
        [DataMember()]
        public string Subscriber;
        [DataMember()]
        public string Classification;
    }

    [ServiceContract(Namespace = "http://schemas.microsoft.com/TeamFoundation/2005/06/Services/Notification/03")]
    interface INotificationService
    {
        [OperationContract(Action = "http://schemas.microsoft.com/TeamFoundation/2005/06/Services/Notification/03/Notify")]
        void Notify(string eventXml, string tfsIdentityXml, SubscriptionInfo SubscriptionInfo);
    }

    public class NotificationService : INotificationService
    {
        private enum NotificationEventType
        {
            NotUsed,
            CheckinEvent,
            BuildCompletionEvent,
            BuildStatusChangeEvent,
            WorkItemChangedEvent
        }

        private NotificationEventType DetermineEventType(string xml)
        {
            if (xml.Contains("<CheckinEvent"))
                return NotificationEventType.CheckinEvent;
            else if (xml.Contains("<BuildCompletionEvent"))
                return NotificationEventType.BuildCompletionEvent;
            else if (xml.Contains("<BuildStatusChangeEvent"))
                return NotificationEventType.BuildStatusChangeEvent;
            else if (xml.Contains("<WorkItemChangedEvent"))
                return NotificationEventType.WorkItemChangedEvent;
            else
                return NotificationEventType.NotUsed;
        }

        public void Notify(string eventXml, string tfsIdentityXml, SubscriptionInfo subscriptionInfo)
        {
            // Handle notification based on type of subscription.
            switch (DetermineEventType(eventXml))
            {
                case NotificationEventType.CheckinEvent:
                    Trace.TraceInformation("{0} NotificationService - Started processing of CheckinEvent", DateTime.Now);
                    CheckinEvent checkinEvent = Common.Deserialize<CheckinEvent>(eventXml);
                    TriggerBuilds(checkinEvent);
                    AdministrationService.CallClients(SubscriptionService.EventType.CheckinEvent, eventXml);
                    Trace.TraceInformation("{0} NotificationService - Finnished processing of CheckinEvent", DateTime.Now);
                    break;

                case NotificationEventType.WorkItemChangedEvent:
                    Trace.TraceInformation("{0} NotificationService - Started processing of WorkItemChangedEvent", DateTime.Now);
                    WorkItemChangedEvent workItemChangedEvent = Common.Deserialize<WorkItemChangedEvent>(eventXml);
                    AdministrationService.CallClients(SubscriptionService.EventType.WorkItemChangedEvent, eventXml);
                    Trace.TraceInformation("{0} NotificationService - Finnished processing of WorkItemChangedEvent", DateTime.Now);
                    break;

                case NotificationEventType.BuildCompletionEvent:
                    Trace.TraceInformation("{0} NotificationService - Started processing of BuildCompleteEvent", DateTime.Now);
                    BuildCompletionEvent buildCompletionEvent = Common.Deserialize<BuildCompletionEvent>(eventXml);

                    //Get the build details
                    TeamBuild build = TFS.BuildService.LoadBuild(buildCompletionEvent.TeamProject, buildCompletionEvent.Id);

                    //Load the build type information
                    BuildType buildType = TFS.BuildService.LoadBuildType(buildCompletionEvent.TeamProject, build.BuildType);

                    //Find registered QueueEntry for build
                    QueueEntry entry = BuildManager.UnregisterBuildFromNotification(build.BuildUri);

                    if (buildType != null && build != null && entry != null)
                    {
                        //Determine if build is a CI build
                        if (entry.TriggeringChangesetID.HasValue)
                        {
                            if (build.Status == BuildStatus.BuildFailed)
                            {
                                TriggerSendingNotifcationEmail(buildCompletionEvent, build, buildType, entry);
                                TriggerCreatingWorkItem(buildCompletionEvent, build, buildType, entry);
                                BuildLog.WriteEntryToLog(EventType.FailedCheckin, buildType.TeamProjectName, buildType.Name, entry.TriggeringUser, build.BuildUri);
                            }
                            else
                            {
                                BuildLog.WriteEntryToLog(EventType.CompletedCheckin, buildType.TeamProjectName, buildType.Name, entry.TriggeringUser, build.BuildUri);
                            }
                        }
                        else
                        {
                            if (build.Status == BuildStatus.BuildFailed)
                            {
                                BuildLog.WriteEntryToLog(EventType.FailedScheduled, buildType.TeamProjectName, buildType.Name, "", build.BuildUri);
                            }
                            else
                            {
                                BuildLog.WriteEntryToLog(EventType.CompletedScheduled, buildType.TeamProjectName, buildType.Name, "", build.BuildUri);
                            }
                        }
                    }

                    if (buildType != null && build != null)
                    {
                        TriggerSendingCompletionEmail(buildCompletionEvent, build, buildType);
                    }

                    AdministrationService.CallClients(SubscriptionService.EventType.BuildCompletionEvent, eventXml);
                    Trace.TraceInformation("{0} NotificationService - Finnished processing of BuildCompleteEvent", DateTime.Now);
                    break;

                case NotificationEventType.BuildStatusChangeEvent:
                    Trace.TraceInformation("{0} NotificationService - Started processing of BuildStatusChangeEvent", DateTime.Now);
                    BuildStatusChangeEvent buildStatusChangeEvent = Common.Deserialize<BuildStatusChangeEvent>(eventXml);
                    AdministrationService.CallClients(SubscriptionService.EventType.BuildStatusChangeEvent, eventXml);
                    Trace.TraceInformation("{0} NotificationService - Finnished processing of BuildStatusChangeEvent", DateTime.Now);
                    break;
                default:
                    Trace.TraceInformation("{0} NotificationService - Unhandled event received. Id: {1}", DateTime.Now, subscriptionInfo.ID);
                    break;
            }

            Trace.TraceInformation("{0} NotificationService - Processed event {1}.", DateTime.Now, eventXml);
        }

        private void TriggerCreatingWorkItem(BuildCompletionEvent buildCompletionEvent, TeamBuild build, BuildType buildType, QueueEntry entry)
        {
            try
            {
                Microsoft.TeamFoundation.Server.Identity owner = TFS.GetUserInfo(entry.TriggeringUser);
                string title = "Changeset [" + entry.TriggeringChangesetID.Value + "] has broken the build [" + build.BuildNumber + "] please investigate";
                string description = "This work item was created by TfsBuildLab on a continous integration build failure.";
                string witTypeToCreate = ConfigurationManager.AppSettings["CINotificationWorkItemType"];
                string witFieldsToSet = ConfigurationManager.AppSettings["CINotificationWorkItemFields"];

                if (owner != null && !String.IsNullOrEmpty(witTypeToCreate))
                {
                    Trace.TraceInformation("{0} NotificationService - Begun creation of work item", DateTime.Now);

                    Dictionary<string, string> fields = new Dictionary<string, string>();
                    fields.Add("System.AssignedTo", owner.DisplayName);

                    if (!String.IsNullOrEmpty(witFieldsToSet))
                    {
                        string[] tempNameValuePair = witFieldsToSet.Split(';');
                        for (int i = 0; i < tempNameValuePair.Length; i++)
                        {
                            string[] tempFieldData = tempNameValuePair[i].Split('=');
                            fields.Add(tempFieldData[0], tempFieldData[1]);
                        }
                    }

                    TFS.CreateWorkItem(buildCompletionEvent.TeamProject, witTypeToCreate, title, description, fields);

                    Trace.TraceInformation("{0} NotificationService - Done creating work item", DateTime.Now);
                }
            }
            catch (Exception ex)
            {
                BuildLog.WriteEntryToLog(EventType.Error, buildCompletionEvent.TeamProject, buildCompletionEvent.Configuration, "", buildCompletionEvent.Url, ex.ToString());
            }
        }

        private void TriggerSendingNotifcationEmail(BuildCompletionEvent buildCompletionEvent, TeamBuild build, BuildType buildType, QueueEntry entry)
        {
            try
            {
                string smtpServerAddress = ConfigurationManager.AppSettings["smtpServerAddress"];
                string administrativeEmail = ConfigurationManager.AppSettings["CIAdministrativeNotificationEmail"];
                string senderEmail = ConfigurationManager.AppSettings["CINotificationSenderAddress"];

                if (String.IsNullOrEmpty(smtpServerAddress))
                {
                    return;
                }

                Trace.TraceInformation("{0} NotificationService - Begin sending of notification email(s)", DateTime.Now);

                SmtpClient smtpClient = new SmtpClient(smtpServerAddress);

                string body = @"You can find the build report here:\n";
                body += @"\t" + buildCompletionEvent.Url + @"\n";
                body += @"For a detailed log of the build please check here:\n";
                body += @"\t" + build.LogLocation + @"\n";

                if (entry.InvolvedChangesets.Count > 1)
                {
                    body += "\n\n" + "These changesets may also be involved in breaking the build:\n";
                    foreach (Changeset potentialCulprits in entry.InvolvedChangesets)
                    {
                        body += "\t" + "Changeset [" + potentialCulprits.Id + "] committed by [" + potentialCulprits.Comitter + "]\n";
                    }
                }

                foreach (Changeset includedChangeset in entry.InvolvedChangesets)
                {
                    string subject = "";
                    if (entry.InvolvedChangesets.Count > 1)
                    {
                        subject = "Your Changeset [" + includedChangeset.Id + "] may have broken the build [" + build.BuildNumber + "]";
                    }
                    else
                    {
                        subject = "Your Changeset [" + includedChangeset.Id + "] has broken the build [" + build.BuildNumber + "]";
                    }

                    Microsoft.TeamFoundation.Server.Identity comitter = TFS.GetUserInfo(includedChangeset.Comitter);
                    if (comitter != null)
                    {
                        string comitterEmail = comitter.MailAddress;
                        if (comitterEmail != String.Empty)
                        {
                            smtpClient.Send(senderEmail, comitterEmail, subject, body);
                        }
                    }
                }

                if (administrativeEmail != "")
                {
                    string subject = "";
                    if (entry.InvolvedChangesets.Count > 1)
                    {
                        subject = entry.TriggeringUser + " checked in changeset [" + entry.TriggeringChangesetID + "] which may be involved in breaking the build [" + build.BuildNumber + "]";
                    }
                    else
                    {
                        subject = entry.TriggeringUser + " checked in changeset [" + entry.TriggeringChangesetID + "] that broke the build [" + build.BuildNumber + "]";
                    }

                    smtpClient.Send(senderEmail, administrativeEmail, subject, body);
                }

                Trace.TraceInformation("{0} NotificationService - Done sending notification email(s)", DateTime.Now);
            }
            catch (Exception ex)
            {
                BuildLog.WriteEntryToLog(EventType.Error, buildCompletionEvent.TeamProject, buildCompletionEvent.Configuration, "", buildCompletionEvent.Url, ex.ToString());
            }
        }
        
        private void TriggerSendingCompletionEmail(BuildCompletionEvent buildCompletionEvent, TeamBuild build, BuildType buildType)
        {
            try
            {
                string smtpServerAddress = ConfigurationManager.AppSettings["smtpServerAddress"];
                string administrativeEmail = ConfigurationManager.AppSettings["CIBuildCompletionNotificationEmail"];
                string senderEmail = ConfigurationManager.AppSettings["CINotificationSenderAddress"];

                if (String.IsNullOrEmpty(smtpServerAddress))
                {
                    return;
                }

                Trace.TraceInformation("{0} NotificationService - Begin sending of completion status email(s)", DateTime.Now);

                SmtpClient smtpClient = new SmtpClient(smtpServerAddress);
               
                string subject = buildType.TeamProjectName + " - " + buildType.Name + " has completed with the status: " + buildCompletionEvent.CompletionStatus;

                string body = @"You can find the build report here:\n";
                body += @"\t" + buildCompletionEvent.Url + @"\n";
                body += @"For a detailed log of the build please check here:\n";
                body += @"\t" + build.LogLocation + @"\n";

                if (administrativeEmail != "")
                {
                    smtpClient.Send(senderEmail, administrativeEmail, subject, body);
                }

                Trace.TraceInformation("{0} NotificationService - Done sending completion status email(s)", DateTime.Now);
            }
            catch (Exception ex)
            {
                BuildLog.WriteEntryToLog(EventType.Error, buildCompletionEvent.TeamProject, buildCompletionEvent.Configuration, "", buildCompletionEvent.Url, ex.ToString());
            }
        }

        private static void TriggerBuilds(CheckinEvent checkinEvent)
        {
            try
            {
                // Get changeset from checkin event
                Changeset cs = Changeset.CreateFromEvent(checkinEvent);

                // Get matching team builds for CS items
                List<BuildType> affectedBuildTypes = TFS.BuildService.GetAffectedBuildTypes(cs);

                // Queue up build
                IBuildQueueRepository db = new BuildQueueRepository();
                foreach (BuildType bt in affectedBuildTypes)
                {
                    if (bt.HasMatchingTrigger(cs))
                    {
                        db.PushBuildToQueue(new QueueEntry(bt.TeamProjectName, bt.Name, bt.DefaultBuildServer, cs.Comitter, cs.Id, bt.GetBuildParametersForMatchingTriggers(cs)));
                    }
                }
            }
            catch (Exception ex)
            {
                BuildLog.WriteEntryToLog(EventType.Error, checkinEvent.TeamProject, "", checkinEvent.Committer, "", ex.ToString());
            }
        }
    }
}
