using System;
using System.Diagnostics;
using System.IO;
using System.Xml;
using System.Xml.Serialization;
using System.Text.RegularExpressions;
using System.Drawing;
using System.Threading;
using System.Windows.Forms;
using System.ServiceModel;
using System.Configuration;
using System.Collections.Generic;
using System.Reflection;
using System.DirectoryServices;
using Callista.BuildLab.Core;
using Callista.BuildLab.Core.Model;

namespace Callista.BuildLab.NotifyClient
{
    public enum NotificationType
    {
        Build,
        Checkin
    }

    public class Notification
    {
        private string _id;
        private string _project;
        private NotificationType _type;
        private string _data;

        public string ID
        {
            get { return _id; }
            set { _id = value; }
        }

        public string Project
        {
            get { return _project; }
            set { _project = value; }
        }
        public NotificationType Type
        {
            get { return _type; }
            set { _type = value; }
        }
        
        public string Data
        {
            get { return _data; }
            set { _data = value; }
        }

        public Notification()
        {
        }

        public Notification(string id, string project, NotificationType type, string data)
        {
            _id = id;
            _project = project;
            _type = type;
            _data = data;
        }

        public bool IsMatch(SubscriptionService.EventType eventType, string eventXml)
        {
            switch (Type)
            {
                case NotificationType.Build:
                    if (eventType == SubscriptionService.EventType.BuildCompletionEvent)
                    {
                        BuildCompletionEvent bce = Common.Deserialize<BuildCompletionEvent>(eventXml);
                        if (bce.TeamProject.Equals(this.Project) && bce.Title.Contains(this.Data))
                        {
                            return true;
                        }
                    }
                    break;
                case NotificationType.Checkin:
                    if (eventType == SubscriptionService.EventType.CheckinEvent)
                    {
                        CheckinEvent ce = Common.Deserialize<CheckinEvent>(eventXml);
                        Changeset cs = Changeset.CreateFromEvent(ce);
                        if (ce.TeamProject.Equals(this.Project))
                        {
                            foreach (VersionedItem vi in cs.VersionedItems)
                            {
                                if (Regex.IsMatch(vi.FullPath, this.Data))
                                {
                                    return true;
                                }
                            }
                        }
                    }
                    break;
            }

            return false;
        }

        public string ExtractAndFormatNotification(SubscriptionService.EventType eventType, string eventXml)
        {
            string message = "";

            switch (eventType)
            {
                case SubscriptionService.EventType.BuildCompletionEvent:
                    BuildCompletionEvent bce = Common.Deserialize<BuildCompletionEvent>(eventXml);
                    message += bce.BuildMachine + " finnished building " + bce.Configuration + " for team project " + bce.TeamProject;
                    message += " with the completion status of [ " + bce.CompletionStatus + " ] the uri for the build is: \n";
                    message += "   " + bce.Url; 
                    break;
                case SubscriptionService.EventType.CheckinEvent:
                    CheckinEvent ce = Common.Deserialize<CheckinEvent>(eventXml);
                    int count = ce.Artifacts.Length - 1;
                    message += ce.Committer + " just checkedin changeset number " + ce.Number + " in team project " + ce.TeamProject + "\n\n";
                    message += "Comment: \n"; 
                    message += "   " + ce.Comment + "\n\n";
                    message += "Containing ( " + count + " items ): ";
                    
                    //Skip the first artifact representing the changeset itself...
                    for (int i = 1; i < ce.Artifacts.Length && i <= 10; i++)
                    {
                        message += "\n   " + ce.Artifacts[i].ServerItem;
                    }
                    break;
            }

            return message;
        }
    }

    public class Program : ApplicationContext, IEventNotificationCallback
    {
        #region Static Members

        private static List<Notification> _notifications = new List<Notification>();
        public static List<Notification> RegisteredNotifications
        {
            get { return _notifications; }
        }

        private static SortedList<string, TeamBuild> _monitoredBuilds = new SortedList<string, TeamBuild>();
        public static SortedList<string, TeamBuild> MonitoredBuilds
        {
            get { return _monitoredBuilds; }
        }

        private static Queue<CheckinEvent> _latestCheckinEvents = new Queue<CheckinEvent>();
        public static Queue<CheckinEvent> LatestCheckinEvents
        {
            get { return _latestCheckinEvents; }
        }

        private static Queue<BuildCompletionEvent> _latestBuildEvents = new Queue<BuildCompletionEvent>();
        public static Queue<BuildCompletionEvent> LatestBuildEvents
        {
            get { return _latestBuildEvents; }
        }
        
        #endregion

        #region Private Members

        private System.ComponentModel.IContainer components;
        
        private System.Windows.Forms.NotifyIcon tfsBuildLabNotifyIcon;
        private System.Windows.Forms.ContextMenu tfsBuildLabNotifyIconContextMenu;
        private System.Windows.Forms.MenuItem configContextMenuItem;
        private System.Windows.Forms.MenuItem statusContextMenuItem;
        private System.Windows.Forms.MenuItem resetContextMenuItem;
        private System.Windows.Forms.MenuItem aboutContextMenuItem;
        private System.Windows.Forms.MenuItem exitContextMenuItem;
        
        private System.Windows.Forms.Form configForm = null;
        private System.Windows.Forms.Form statusForm = null;
        private System.Windows.Forms.Form aboutForm = null;

        private System.Timers.Timer _refreshTimer = null;

        private bool _refreshInProgress = false;
        private bool _buildInProgress = false;
        private bool _isAlerting = false;

        #endregion

        #region Initialization 

        public Program()
        {
            if (ConnectToServer())
            {
                InitializeContext();
                InitializeData();
            }
        }

        private void InitializeContext()
        {
            this.components = new System.ComponentModel.Container();
            this.tfsBuildLabNotifyIcon = new System.Windows.Forms.NotifyIcon(this.components);
            this.tfsBuildLabNotifyIconContextMenu = new System.Windows.Forms.ContextMenu();
            this.configContextMenuItem = new System.Windows.Forms.MenuItem();
            this.statusContextMenuItem = new System.Windows.Forms.MenuItem();
            this.resetContextMenuItem = new System.Windows.Forms.MenuItem();
            this.aboutContextMenuItem = new System.Windows.Forms.MenuItem();
            this.exitContextMenuItem = new System.Windows.Forms.MenuItem();

            // 
            // tfsBuildLabNotifyIcon
            // 
            this.tfsBuildLabNotifyIcon.ContextMenu = tfsBuildLabNotifyIconContextMenu;
            this.tfsBuildLabNotifyIcon.DoubleClick += new EventHandler(tfsBuildLabNotifyIcon_DoubleClick);
            this.tfsBuildLabNotifyIcon.Icon = new Icon(typeof(Program), "AppGreen.ico");
            this.tfsBuildLabNotifyIcon.Text = "TfsBuildLab Event Toaster";
            this.tfsBuildLabNotifyIcon.Visible = true;
            
            // 
            // tfsBuildLabNotifyIconContextMenu
            // 
            this.tfsBuildLabNotifyIconContextMenu.MenuItems.AddRange(new MenuItem[] { configContextMenuItem, statusContextMenuItem, resetContextMenuItem, aboutContextMenuItem, exitContextMenuItem });

            // 
            // configContextMenuItem
            // 
            this.configContextMenuItem.Index = 0;
            this.configContextMenuItem.Text = "&Configure";
            this.configContextMenuItem.Click += new System.EventHandler(this.configContextMenuItem_Click);

            // 
            // statusContextMenuItem
            // 
            this.statusContextMenuItem.Index = 1;
            this.statusContextMenuItem.Text = "&Status";
            this.statusContextMenuItem.DefaultItem = true;
            this.statusContextMenuItem.Click += new System.EventHandler(this.statusContextMenuItem_Click);

            // 
            // resetContextMenuItem
            // 
            this.resetContextMenuItem.Index = 2;
            this.resetContextMenuItem.Text = "&Reset alert";
            this.resetContextMenuItem.Click += new System.EventHandler(this.resetContextMenuItem_Click);

            // 
            // aboutContextMenuItem
            // 
            this.aboutContextMenuItem.Index = 3;
            this.aboutContextMenuItem.Text = "&About";
            this.aboutContextMenuItem.Click += new System.EventHandler(this.aboutContextMenuItem_Click);

            // 
            // exitContextMenuItem
            // 
            this.exitContextMenuItem.Index = 4;
            this.exitContextMenuItem.Text = "&Exit";
            this.exitContextMenuItem.Click += new System.EventHandler(this.exitContextMenuItem_Click);
        }

        private void InitializeData()
        {
            string path = Assembly.GetExecutingAssembly().CodeBase.Replace(@"file:///", "");
            path = path.Substring(0,path.LastIndexOf(".")) + ".xml";

            if (File.Exists(path))
            {
                XmlSerializer serializer = new XmlSerializer(typeof(List<Notification>));
                using (TextReader reader = new StreamReader(path))
                {
                    RegisteredNotifications.AddRange((List<Notification>)serializer.Deserialize(reader));
                }
            }

            RefreshMonitoredBuilds(true);
            RefreshStatus(true);
        }

        private static void SerializeData()
        {
            string path = Assembly.GetExecutingAssembly().CodeBase.Replace(@"file:///", "");
            path = path.Substring(0, path.LastIndexOf(".")) + ".xml";

            XmlSerializer serializer = new XmlSerializer(typeof(List<Notification>));
            using (TextWriter writer = new StreamWriter(path, false))
            {
                serializer.Serialize(writer, _notifications);
            }
        }

        #endregion

        #region Shutdown

        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (components != null)
                {
                    components.Dispose();
                }
            }
        }

        protected override void ExitThreadCore()
        {
            if (configForm != null)
                configForm.Close();

            if (statusForm != null)
                statusForm.Close();

            if (aboutForm != null)
                aboutForm.Close();

            DisconnectFromServer();

            base.ExitThreadCore();
        }

        #endregion

        #region Context Menu

        private void configContextMenuItem_Click(object sender, EventArgs e)
        {
            if (aboutForm != null)
            {
                aboutForm.Close();
                aboutForm = null;
            }

            if (statusForm != null)
            {
                statusForm.Close();
                statusForm = null;
            }

            if (configForm == null)
            {
                List<TeamProject> projects = null;
                IBuildTypeAdministration proxy = null;
                try
                {
                    proxy = CreateProxy<IBuildTypeAdministration>();
                    projects = proxy.ListTeamProjects();
                }
                finally
                {
                    CloseProxy((ICommunicationObject)proxy);
                }

                configForm = new ConfigForm(projects);
                configForm.ShowInTaskbar = false;
                configForm.Closed += new EventHandler(ConfigForm_FormClosed);
            }

            configForm.Show();
            configForm.Activate();
        }

        private void statusContextMenuItem_Click(object sender, EventArgs e)
        {
            if (aboutForm != null)
            {
                aboutForm.Close();
                aboutForm = null;
            }

            if (configForm != null)
            {
                configForm.Close();
                configForm = null;
            }

            if (statusForm == null)
            {
                RefreshMonitoredBuilds(true);
                statusForm = new StatusForm();
                statusForm.ShowInTaskbar = false;
                statusForm.Closed += new EventHandler(StatusForm_FormClosed);
            }

            statusForm.Show();
            statusForm.Activate();
        }

        private void resetContextMenuItem_Click(object sender, EventArgs e)
        {   
            this.ResetAlert();
        }

        public void Alert()
        {
            Alert(ConfigurationManager.AppSettings["AlertCommand"]);
        }

        public void Alert(string command)
        {
            ExecuteCommand(command);
            _isAlerting = true;
        }

        public void ResetAlert()
        {
            ResetAlert(ConfigurationManager.AppSettings["ResetCommand"]);
        }

        public void ResetAlert(string command)
        {
            ExecuteCommand(command);
            _isAlerting = false;
        }

        private void ExecuteCommand(string command)
        {
            if (!string.IsNullOrEmpty(command))
            {
                int pos = command.IndexOf(".exe") + 4;
                string fileName = command.Substring(0, pos);
                string arguments = command.Substring(pos + 1);
                Process.Start(fileName, arguments);
            }
        }

        private void aboutContextMenuItem_Click(object sender, EventArgs e)
        {
            if (statusForm != null)
            {
                statusForm.Close();
                statusForm = null;
            }

            if (configForm != null)
            {
                configForm.Close();
                configForm = null;
            }

            if (aboutForm == null)
            {
                aboutForm = new AboutBox();
                aboutForm.ShowInTaskbar = false;
                aboutForm.Closed += new EventHandler(AboutForm_FormClosed);
            }

            aboutForm.Show();
            aboutForm.Activate();
        }

        private void exitContextMenuItem_Click(object sender, EventArgs e)
        {
            SerializeData();
            ExitThread();
        }
        
        #endregion

        #region Event Handlers

        private void ConfigForm_FormClosed(object sender, EventArgs e)
        {
            configForm = null;
            SerializeData();
            RefreshMonitoredBuilds(true);
            RefreshStatus(true);
        }

        private void StatusForm_FormClosed(object sender, EventArgs e)
        {
            statusForm = null;
        }

        private void AboutForm_FormClosed(object sender, EventArgs e)
        {
            aboutForm = null;
        }

        private void tfsBuildLabNotifyIcon_DoubleClick(object sender, System.EventArgs e)
        {
            if (aboutForm != null)
            {
                aboutForm.Close();
                aboutForm = null;
            }

            if (configForm != null)
            {
                configForm.Close();
                configForm = null;
            }

            if (statusForm == null)
            {
                RefreshMonitoredBuilds(true);
                statusForm = new StatusForm();
                statusForm.ShowInTaskbar = false;
                statusForm.Closed += new EventHandler(StatusForm_FormClosed);
            }

            statusForm.Show();
            statusForm.Activate();
        }

        #endregion

        #region Build Status Refresh

        private void RefreshStatusEvent(object sender, System.Timers.ElapsedEventArgs e) 
        {
            if (!_refreshInProgress)
            {
                _refreshInProgress = true;

                if (!_buildInProgress)
                {
                    RefreshMonitoredBuilds(false);
                    RefreshStatus(true);
                }

                _refreshInProgress = false;
            }
        }

        private void RefreshMonitoredBuilds(bool SupressNotifications)
        {
            foreach (Notification n in Program.RegisteredNotifications)
            {
                if (n.Type == NotificationType.Build)
                {
                    TeamBuild latestBuild = null;
                    List<TeamBuild> builds = null;
                    IBuildRepositoryManagement proxy = null;
                    try
                    {
                        proxy = CreateProxy<IBuildRepositoryManagement>();
                        builds = proxy.ListBuilds(n.Project, n.Data, "");
                    }
                    finally
                    {
                        CloseProxy((ICommunicationObject)proxy);
                    }

                    foreach (TeamBuild build in builds)
                    {
                        if (build.Status == BuildStatus.BuildRunning)
                        {
                            latestBuild = build;
                            break;
                        }
                        else
                        {
                            if (latestBuild == null)
                            {
                                latestBuild = build;
                            }
                            else
                            {
                                if (build.FinishTime > latestBuild.FinishTime)
                                {
                                    latestBuild = build;
                                }
                            }
                        }
                    }

                    if (latestBuild != null)
                    {
                        if (_monitoredBuilds.ContainsKey(n.Project + ":" + latestBuild.BuildType))
                        {
                            _monitoredBuilds.Remove(n.Project + ":" + latestBuild.BuildType);
                        }
                       
                        _monitoredBuilds.Add(n.Project + ":" + latestBuild.BuildType, latestBuild);

                        if (latestBuild.Status == BuildStatus.BuildRunning)
                        {
                            if (!SupressNotifications)
                            {
                                string message = latestBuild.BuildMachine + "Started building " + latestBuild.BuildType + " for team project " + n.Project + " the uri for the build is: " + latestBuild.BuildNumber;
                                tfsBuildLabNotifyIcon.ShowBalloonTip(1, "TFS Build", message, ToolTipIcon.Info);
                            }
                            _buildInProgress = true;
                        }
                    }
                }
            }
        }

        private void RefreshStatus(bool suppressAlerts)
        {
            foreach (TeamBuild b in _monitoredBuilds.Values)
            {
                switch(b.Status)
                {
                    case BuildStatus.BuildFailed:
                        if (!suppressAlerts)
                        {
                            this.Alert();
                        }
                        tfsBuildLabNotifyIcon.Icon = new Icon(typeof(Program), "AppRed.ico");
                        return;
                    case BuildStatus.BuildRunning:
                        tfsBuildLabNotifyIcon.Icon = new Icon(typeof(Program), "AppYellow.ico");
                        return;
                }
            }
            // Reset alert if all builds are ok
            if (_isAlerting)
            {
                this.ResetAlert();
            }

            tfsBuildLabNotifyIcon.Icon = new Icon(typeof(Program), "AppGreen.ico");
        }

        #endregion

        #region TfsBuildLab Proxy Handling

        private IAdministration _adminProxy = null;
        delegate void ShutdownApplicationViaWorkerThread();

        private string TfsBuildLabServer
        {
            get
            {
                string tfsBuildLab = "";

                if (!string.IsNullOrEmpty(ConfigurationManager.AppSettings["TFSBuildLab"]))
                {
                    tfsBuildLab = ConfigurationManager.AppSettings["TFSBuildLab"];
                }

                return tfsBuildLab;
            }
        }

        private NetTcpBinding CreateBinding()
        {
            NetTcpBinding binding = new NetTcpBinding(SecurityMode.None);

            binding.ReaderQuotas.MaxArrayLength = int.MaxValue;
            binding.ReaderQuotas.MaxBytesPerRead = int.MaxValue;
            binding.ReaderQuotas.MaxDepth = int.MaxValue;
            binding.ReaderQuotas.MaxStringContentLength = int.MaxValue;
            binding.ReaderQuotas.MaxBytesPerRead = int.MaxValue;

            binding.MaxReceivedMessageSize = int.MaxValue;
            binding.MaxBufferSize = int.MaxValue;

            binding.ReliableSession.Enabled = false;

            return binding;
        }

        private T CreateProxy<T>()
        {
            T proxy = default(T);
            try
            {
                string path = string.Format("net.tcp://{0}/", TfsBuildLabServer);
                ChannelFactory<T> factory = new ChannelFactory<T>(CreateBinding(), path);
                factory.Endpoint.Behaviors.Add(new System.ServiceModel.Description.ClientViaBehavior(new Uri(path)));
                proxy = factory.CreateChannel();
                ICommunicationObject co = (ICommunicationObject)proxy;
                co.Open();
            }
            catch (Exception e)
            {
                MessageBox.Show(string.Format("Failed to connect to TFSBuildLab service on server '{0}'. Reason: {1}", TfsBuildLabServer, e.Message), "Connection failed closing down client!!!", MessageBoxButtons.OK);
                ExitThreadCore();
            }

            return proxy;
        }

        private void CloseProxy(ICommunicationObject co)
        {
            if (co != null)
            {
                if (co.State == CommunicationState.Opened)
                {
                    co.Close();
                }
            }
        }

        private bool ConnectToServer()
        {
            try
            {
                string path = string.Format("net.tcp://{0}/", TfsBuildLabServer);
                string frequency = ConfigurationManager.AppSettings["RefreshFrequency"];

                DuplexChannelFactory<IAdministration> adminFactory = new DuplexChannelFactory<IAdministration>(new InstanceContext(this), CreateBinding(), path);
                CallbackBehaviorAttribute callbackBehaviour = (CallbackBehaviorAttribute)adminFactory.Endpoint.Behaviors[typeof(CallbackBehaviorAttribute)];
                callbackBehaviour.ConcurrencyMode = ConcurrencyMode.Reentrant;

                _adminProxy = adminFactory.CreateChannel();
                _adminProxy.Connect();

                ICommunicationObject co = (ICommunicationObject)_adminProxy;
                co.Faulted += new EventHandler(ServerConnectionBroke);

                if (String.IsNullOrEmpty(frequency))
                    frequency = "5000";

                _refreshTimer = new System.Timers.Timer();
                _refreshTimer.Interval = Int32.Parse(frequency);
                _refreshTimer.Elapsed += new System.Timers.ElapsedEventHandler(this.RefreshStatusEvent);
                _refreshTimer.Enabled = true;
            }
            catch (Exception e)
            {
                _adminProxy = null;
                if (_refreshTimer != null)
                {
                    _refreshTimer.Close();
                    _refreshTimer = null;
                }
                MessageBox.Show(string.Format("Failed to connect to TFSBuildLab service on server '{0}'. Reason: {1}", TfsBuildLabServer, e.Message), "Connection failed closing down client!!!", MessageBoxButtons.OK);
                ExitThreadCore();
                return false;
            }

            return true;
        }

        private void DisconnectFromServer()
        {
            if (_adminProxy != null)
            {
                _adminProxy.Disconnect();
                CloseProxy((ICommunicationObject)_adminProxy);
            }
        }

        void ServerConnectionBroke(object sender, EventArgs e)
        {
            _adminProxy = null;
            if (_refreshTimer != null)
            {
                _refreshTimer.Close();
                _refreshTimer = null;
            }

            if (MessageBox.Show(string.Format("Connect to TFSBuildLab service on server '{0}' was lost do you wish to try and reconnect?", TfsBuildLabServer), "Server connection unexpectedly faulted!!!", MessageBoxButtons.YesNo) == DialogResult.Yes)
            {
                ConnectToServer();
            }
            else
            {
                ExitThreadCore();
            }
        }

        #endregion

        #region IEventNotificationCallback Members

        void IEventNotificationCallback.Notify(SubscriptionService.EventType eventType, string eventXml)
        {
            if (eventType == SubscriptionService.EventType.WorkItemChangedEvent)
            {
                DirectoryEntry userEntry = new DirectoryEntry("WinNT://" + Environment.UserDomainName + "/" + Environment.UserName + ",User");

                string message = "";
                string name = (string)userEntry.Properties["fullname"].Value;
                string id = "";

                WorkItemChangedEvent workItemChangedEvent = Common.Deserialize<WorkItemChangedEvent>(eventXml);

                if (workItemChangedEvent.ChangedFields != null && workItemChangedEvent.ChangedFields.StringFields != null)
                {
                    if (workItemChangedEvent.CoreFields != null && workItemChangedEvent.CoreFields.IntegerFields != null)
                    {
                        foreach (IntegerField field in workItemChangedEvent.CoreFields.IntegerFields)
                        {
                            if (field.ReferenceName == "System.Id")
                            {
                                id = field.NewValue.ToString();
                            }
                        }
                    }

                    foreach (StringField field in workItemChangedEvent.ChangedFields.StringFields)
                    {
                        if (field.ReferenceName == "System.AssignedTo")
                        {
                            string oldValue = string.IsNullOrEmpty(field.OldValue) ? "" : field.OldValue;
                            string newValue = string.IsNullOrEmpty(field.NewValue) ? "" : field.NewValue;

                            if (name != newValue && name == oldValue)
                            {
                                message += "You are no longer assigned to working on the following work item:";
                                message += "\n\t" + workItemChangedEvent.Title;
                                message += "\n\n" + workItemChangedEvent.DisplayUrl;

                                tfsBuildLabNotifyIcon.ShowBalloonTip(1, "TFS Work Item Assignment Removal [" + id + "]", message, ToolTipIcon.Info);
                            }

                            if (name == newValue && name != oldValue)
                            {
                                message += "You have been assigned to work on the following work item:";
                                message += "\n\t" + workItemChangedEvent.Title;
                                message += "\n\n" + workItemChangedEvent.DisplayUrl;

                                tfsBuildLabNotifyIcon.ShowBalloonTip(1, "TFS Work Item Assignment [" + id + "]", message, ToolTipIcon.Info);
                            }

                            return;
                        }
                    }
                }

                return; //No match on "Assigned To" skip this WIT
            }

            foreach (Notification n in Program.RegisteredNotifications)
            {
                if (n.IsMatch(eventType, eventXml))
                {
                    string message = n.ExtractAndFormatNotification(eventType, eventXml);
                    switch (eventType)
                    {
                        case SubscriptionService.EventType.CheckinEvent:
                            tfsBuildLabNotifyIcon.ShowBalloonTip(1, "TFS Checkin", message, ToolTipIcon.Info);

                            CheckinEvent ce = Common.Deserialize<CheckinEvent>(eventXml);
                            if (_latestCheckinEvents.Count == 10)
                            {
                                _latestCheckinEvents.Dequeue();
                            }
                            _latestCheckinEvents.Enqueue(ce);
                            
                            break;
                        case SubscriptionService.EventType.BuildCompletionEvent:
                            tfsBuildLabNotifyIcon.ShowBalloonTip(1, "TFS Build", message, ToolTipIcon.Info);
                            
                            BuildCompletionEvent bce = Common.Deserialize<BuildCompletionEvent>(eventXml);
                            if (_latestBuildEvents.Count == 10)
                            {
                                _latestBuildEvents.Dequeue();
                            }
                            _latestBuildEvents.Enqueue(bce);

                            TeamBuild build = null;
                            IBuildRepositoryManagement proxy = null;
                            try
                            {
                                proxy = CreateProxy<IBuildRepositoryManagement>();
                                build = proxy.GetSpecificBuild(bce.TeamProject, bce.Id);
                            }
                            finally
                            {
                                CloseProxy((ICommunicationObject)proxy);
                            }

                            if (_monitoredBuilds.ContainsKey(bce.TeamProject + ":" + build.BuildType))
                            {
                                _monitoredBuilds.Remove(bce.TeamProject + ":" + build.BuildType);
                            }
                            _monitoredBuilds.Add(bce.TeamProject + ":" + build.BuildType, build);

                            _buildInProgress = false;
                            foreach (TeamBuild tb in _monitoredBuilds.Values)
                            {
                                if (tb.Status == BuildStatus.BuildRunning)
                                {
                                    _buildInProgress = true;
                                }
                            }

                            RefreshStatus(false);

                            break;
                    }
                }
            }
        }

        #endregion

        [STAThread]
        static void Main()
        {
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);
            Application.Run(new Program());
        }
    }
}
