using System;
using System.Windows.Forms;
using System.Collections.Generic;
using System.ServiceModel;
using System.Configuration;
using Microsoft.TeamFoundation.VersionControl.Client;
using Callista.BuildLab.Core;
using Callista.BuildLab.Core.Model;

namespace TfsBuildLab.Policies.BrokenBuild
{
    [Serializable]
    public class BrokenBuildPolicy : PolicyBase
    {
        private IBuildTypeAdministration _buildTypeAdminProxy = null;
        private IBuildRepositoryManagement _buildRepositoryProxy = null;

        private const string instructions = "Deploy the assembly to your machine and create the following entry in the registry \n\n" +
                                            "\t Key: [HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\VisualStudio\\8.0\\TeamFoundation\\SourceControl\\Checkin Policies] \n" +
                                            "\t Name: name of assembly (without extension) \n" +
                                            "\t Value: full path to assembly";

        public override string Description
        {
            get { return "Prevent users from accidental checkins on broken builds"; }
        }

        public override string Type
        {
            get { return "TfsBuildLab Broken Build Policy"; }
        }

        public override string TypeDescription
        {
            get { return "This policy will force the user to make a consious decision when checking in on a broken build."; }
        }

        public override string InstallationInstructions
        {
            get { return instructions; }
        }

        public override bool CanEdit
        {
            get { return false; }
        }

        public override void Activate(PolicyFailure failure)
        {
            MessageBox.Show("If you really need to checkin even though the build has failed you must override this policy.", "How to fix your policy failure");
        }

        public override bool Edit(IPolicyEditArgs policyEditArgs)
        {
            return true;
        }

        public override void DisplayHelp(PolicyFailure failure)
        {
            MessageBox.Show("This policy requires you to make an active choice when dealing with broken builds and must be overridden to perform a checkin.", "Prompt Policy Help");
        }

        public override PolicyFailure[] Evaluate()
        {
            List<PolicyFailure> failures = new List<PolicyFailure>();

            try 
	        {	        
                CreateAdminProxy();

                List<BuildType> buildTypes = _buildTypeAdminProxy.ListBuildTypes(String.Empty);
                foreach (PendingChange change in PendingCheckin.PendingChanges.CheckedPendingChanges)
                {
                    foreach (BuildType buildType in buildTypes)
                    {
                        foreach (string path in buildType.WorkspaceMapping.MappedPaths)
                        {
                            if (change.ServerItem.Contains(path))
                            {
                                TeamBuild latestBuild = _buildRepositoryProxy.GetLatestBuild(buildType.TeamProjectName, buildType.Name, string.Empty);
                                if (latestBuild.Status == BuildStatus.BuildFailed)
                                {
                                    failures.Add(CreatePolicyFailure(buildType, latestBuild, change));
                                }
                            }
                        }
                    }
                }
            }
	        catch (Exception ex)
	        {
                return new PolicyFailure[] { new PolicyFailure("An exception occured ... " + ex.Message, this) };
	        }
            finally 
            {
                CloseAdminProxy();
            }

            return failures.ToArray();
        }

        private PolicyFailure CreatePolicyFailure(BuildType buildType, TeamBuild build, PendingChange change)
        {
            return new PolicyFailure(String.Format("{0} is included in a broken build [{1}]. Please fix the build or notify the party resposible for the build break.", change.ServerItem, build.BuildNumber), this);
        }

        #region AdminProxy Members

        private bool CreateAdminProxy()
        {
            string server = "";
            string path = "";

            try
            {
                NetTcpBinding binding = new NetTcpBinding(SecurityMode.None);

                binding.ReaderQuotas.MaxArrayLength = int.MaxValue;
                binding.ReaderQuotas.MaxBytesPerRead = int.MaxValue;
                binding.ReaderQuotas.MaxDepth = int.MaxValue;
                binding.ReaderQuotas.MaxStringContentLength = int.MaxValue;
                binding.ReaderQuotas.MaxBytesPerRead = int.MaxValue;

                binding.MaxReceivedMessageSize = int.MaxValue;
                binding.MaxBufferSize = int.MaxValue;

                binding.ReliableSession.Enabled = false;

                server = ConfigurationManager.OpenExeConfiguration(typeof(BrokenBuildPolicy).Assembly.Location).AppSettings.Settings["TFSBuildLab"].Value;
                path = string.Format("net.tcp://{0}/", server);

                ChannelFactory<IBuildRepositoryManagement> factoryBuildRepository = new ChannelFactory<IBuildRepositoryManagement>(binding, path);
                _buildRepositoryProxy = factoryBuildRepository.CreateChannel();

                ChannelFactory<IBuildTypeAdministration> factoryBuildTypeAdmin = new ChannelFactory<IBuildTypeAdministration>(binding, path);
                _buildTypeAdminProxy = factoryBuildTypeAdmin.CreateChannel();
            }
            catch (Exception e)
            {
                MessageBox.Show(string.Format("Failed to connect to TFSBuildLab service on server '{0}'. Reason: {1}", server, e.Message), "Connection failed", MessageBoxButtons.OK);
                return false;
            }

            return true;
        }

        private void CloseAdminProxy()
        {
            if (_buildRepositoryProxy != null)
            {
                if (((ICommunicationObject)_buildRepositoryProxy).State == CommunicationState.Opened)
                {
                    ((ICommunicationObject)_buildRepositoryProxy).Close();
                }
                _buildRepositoryProxy = null;
            }

            if (_buildTypeAdminProxy != null)
            {
                if (((ICommunicationObject)_buildTypeAdminProxy).State == CommunicationState.Opened)
                {
                    ((ICommunicationObject)_buildTypeAdminProxy).Close();
                }
                _buildTypeAdminProxy = null;
            }
        }

        #endregion
    }
}
