using System;
using System.Windows.Forms;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using Microsoft.TeamFoundation.VersionControl.Client;

namespace Callista.BuildLab.Policies.RestrictedPaths
{
    [Serializable]
    public class RestrictedPathsPolicy : PolicyBase
    {
        private List<CheckinRestriction> _restrictions = new List<CheckinRestriction>();

        private const string instructions = "Deploy the assembly to your machine and create the following entry in the registry \n\n" +
                                            "\t Key: [HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\VisualStudio\\8.0\\TeamFoundation\\SourceControl\\Checkin Policies] \n" +
                                            "\t Name: name of assembly (without extension) \n" +
                                            "\t Value: full path to assembly";
        public override string Description
        {
            get { return "Prevent users from accidental checkins on restricted paths"; }
        }

        public override string Type
        {
            get { return "TfsBuildLab Restricted Paths Policy"; }
        }

        public override string TypeDescription
        {
            get { return "This policy will force the user to make a consious decision when checking in to a restricted path."; }
        }

        public override string InstallationInstructions
        {
            get { return instructions; }
        }

        public override bool CanEdit
        {
            get { return true; }
        }

        public override void Activate(PolicyFailure failure)
        {
            MessageBox.Show("If you really need to checkin in your restricted resources you must override this policy.", "How to fix your policy failure");
        }

        public override bool Edit(IPolicyEditArgs policyEditArgs)
        {
            RestrictedPathsForm dlg = new RestrictedPathsForm();
            dlg.Restrictions = _restrictions;

            if (policyEditArgs.IsNew)
                return true;

            if (dlg.ShowDialog() == DialogResult.OK)
            {
                _restrictions = dlg.Restrictions;
                return true;
            }
            
            return false;
        }

        public override void DisplayHelp(PolicyFailure failure)
        {
            MessageBox.Show("This policy requires you to make an active choice when dealing with restricted resources and must be overridden to perform a checkin.", "Prompt Policy Help");
        }

        public override PolicyFailure[] Evaluate()
        {
            List<PolicyFailure> failures = new List<PolicyFailure>();

            try 
	        {
                foreach (PendingChange change in PendingCheckin.PendingChanges.CheckedPendingChanges)
                {
                    foreach (CheckinRestriction restriction in _restrictions)
                    {
                        if (Regex.IsMatch(change.ServerItem, restriction.Path))
                        {
                            if (restriction.BreakOnFirstViolation)
                            {
                                return new PolicyFailure[] { CreatePolicyFailure(restriction, change) };
                            }
                            else
                            {
                                failures.Add(CreatePolicyFailure(restriction, change));
                            }
                        }
                    }
                }
	        }
	        catch (Exception ex)
	        {
                return new PolicyFailure[] { new PolicyFailure("An exception occured ... " + ex.Message, this) };
	        }

            return failures.ToArray();
        }

        private PolicyFailure CreatePolicyFailure(CheckinRestriction restriction, PendingChange change)
        {
            return new PolicyFailure(String.Format("{0} is included in the restricted path [{1}]. The reason for the restriction is due to: {2}.", change.ServerItem, restriction.Name, restriction.Comment), this);
        }
    }
}
